"""
Test file input/output.
"""
import os
import unittest
import numpy as np

import ams


class TestMATPOWER(unittest.TestCase):
    """
    Test IO functions for MATPOWER and PYPOWER.
    """

    def setUp(self):
        self.mpc5 = ams.io.matpower.m2mpc(ams.get_case('matpower/case5.m'))
        self.mpc14 = ams.io.matpower.m2mpc(ams.get_case('matpower/case14.m'))

    def test_m2mpc(self):
        """Test conversion from M file to mpc dict."""
        # NOTE: when the keys are there, read them
        self.assertTupleEqual(self.mpc5['gencost'].shape, (5, 6))
        self.assertTupleEqual(self.mpc5['gentype'].shape, (5,))
        self.assertTupleEqual(self.mpc5['genfuel'].shape, (5,))

        # NOTE: when the keys are not there, the read mpc will not complete them
        self.assertTupleEqual(self.mpc14['gencost'].shape, (5, 7))
        self.assertNotIn('gentype', self.mpc14)
        self.assertNotIn('genfuel', self.mpc14)

    def test_mpc2system(self):
        """Test conversion from MPC to AMS System."""
        system5 = ams.system.System()
        ams.io.matpower.mpc2system(self.mpc5, system5)
        # In case5.m, the gencost has type 2 cost model, with 2 parameters.
        np.testing.assert_array_equal(system5.GCost.c2.v,
                                      np.zeros(system5.StaticGen.n))

        # Check if Area is added
        self.assertGreater(system5.Area.n, 0)

        # Check if Zone is added
        self.assertGreater(system5.Zone.n, 0)

        system14 = ams.system.System()
        # Test gentype length check
        mpc14 = self.mpc14.copy()
        mpc14['gentype'] = np.array(['WT'] * 6)
        with self.assertRaises(ValueError, msg='gentype length check failed!'):
            ams.io.matpower.mpc2system(mpc14, system14)

        system14 = ams.system.System()
        ams.io.matpower.mpc2system(self.mpc14, system14)
        # In case14.m, the gencost has type 2 cost model, with 3 parameters.
        np.testing.assert_array_less(np.zeros(system14.StaticGen.n),
                                     system14.GCost.c2.v,)

    def test_system2mpc(self):
        """Test conversion from AMS System to MPC."""
        system5 = ams.system.System()
        ams.io.matpower.mpc2system(self.mpc5, system5)
        mpc5 = ams.io.matpower.system2mpc(system5)

        self.assertEqual(mpc5['baseMVA'], self.mpc5['baseMVA'])

        # Bus
        # type, PD, QD, GS,BS, VM, VA. BASE_KV, VMAX, VMIN
        bus_cols = [1, 2, 3, 4, 5, 7, 8, 9, 11, 12]
        np.testing.assert_array_equal(mpc5['bus'][:, bus_cols],
                                      self.mpc5['bus'][:, bus_cols])

        # Branch, Gen, Gencost, can have minor differences but is okay

        # String type data
        np.testing.assert_array_equal(mpc5['gentype'], self.mpc5['gentype'])
        np.testing.assert_array_equal(mpc5['genfuel'], self.mpc5['genfuel'])
        np.testing.assert_array_equal(mpc5['bus_name'], self.mpc5['bus_name'])

        # Area quantity
        self.assertEqual(np.unique(mpc5['bus'][:, 6]).shape[0],
                         np.unique(self.mpc5['bus'][:, 6]).shape[0])

        # Zone quantity
        self.assertEqual(np.unique(mpc5['bus'][:, 10]).shape[0],
                         np.unique(self.mpc5['bus'][:, 10]).shape[0])

    def test_gencost1(self):
        """Test when gencost is type 1."""
        mpcgc1 = self.mpc14.copy()
        mpcgc1['gencost'] = np.repeat(np.array([[1, 0, 0, 3, 0.01, 40, 0]]), 5, axis=0)

        system = ams.system.System()
        ams.io.matpower.mpc2system(mpcgc1, system)
        self.assertEqual(system.GCost.n, 5)

    def test_mpc2m(self):
        """Test conversion from MPC to M file."""
        mpc5 = ams.io.matpower.m2mpc(ams.get_case('matpower/case5.m'))
        mpc14 = ams.io.matpower.m2mpc(ams.get_case('matpower/case14.m'))

        # Test conversion to M file
        mfile5 = ams.io.matpower.mpc2m(mpc5, './case5out.m')
        mfile14 = ams.io.matpower.mpc2m(mpc14, './case14out.m')

        # Check if the files exist
        self.assertTrue(os.path.exists(mfile5))
        self.assertTrue(os.path.exists(mfile14))

        mpc5read = ams.io.matpower.m2mpc(mfile5)
        mpc14read = ams.io.matpower.m2mpc(mfile14)

        # Check if the numerical values are the same
        for key in mpc5:
            if key in ['bus_name', 'gentype', 'genfuel']:
                continue
            np.testing.assert_array_almost_equal(
                mpc5[key], mpc5read[key], decimal=5,
                err_msg=f"Mismatch in {key} when converting case5.m"
            )
        for key in mpc14:
            if key in ['bus_name', 'gentype', 'genfuel']:
                continue
            np.testing.assert_array_almost_equal(
                mpc14[key], mpc14read[key], decimal=5,
                err_msg=f"Mismatch in {key} when converting case14.m"
            )

        # Clean up the generated files
        os.remove(mfile5)
        os.remove(mfile14)

    def test_system2m(self):
        """Test conversion from AMS System to M file."""
        s5 = ams.load(ams.get_case('matpower/case5.m'),
                      no_output=True)
        mfile5 = './case5out.m'
        ams.io.matpower.write(s5, mfile5)

        # Clean up the generated file
        os.remove(mfile5)


class TestSystemExport(unittest.TestCase):
    """
    Test system export functions.
    """

    def test_system_to(self):
        """Test conversion from AMS System to several formats."""
        ss = ams.load(ams.get_case('matpower/case5.m'),
                      no_output=True)

        mpc = ss.to_mpc()

        self.assertIsInstance(mpc, dict)
        self.assertIn('bus', mpc)
        self.assertIn('branch', mpc)
        self.assertIn('gen', mpc)
        self.assertIn('gencost', mpc)
        self.assertIn('baseMVA', mpc)
        self.assertIn('bus_name', mpc)
        self.assertIn('gentype', mpc)
        self.assertIn('genfuel', mpc)

        ss.to_m("./case5_out.m")
        self.assertTrue(os.path.exists("./case5_out.m"))
        os.remove("./case5_out.m")

        ss.to_json("./case5_out.json")
        self.assertTrue(os.path.exists("./case5_out.json"))
        os.remove("./case5_out.json")

        ss.to_xlsx("./case5_out.xlsx")
        self.assertTrue(os.path.exists("./case5_out.xlsx"))
        os.remove("./case5_out.xlsx")

        ss.to_raw("./case5_out.raw", overwrite=True)
        self.assertTrue(os.path.exists("./case5_out.raw"))
        os.remove("./case5_out.raw")



"""
DCOPF routines.
"""
import logging

import numpy as np
from ams.core.param import RParam
from ams.core.service import NumOp

from ams.routines.dcopf import DCOPF
from ams.opt import ExpressionCalc

from ams.shared import sps


logger = logging.getLogger(__name__)


class DCOPF2(DCOPF):
    """
    DC optimal power flow (DCOPF) using PTDF.
    For large cases, it is recommended to build the PTDF first, especially when incremental
    build is necessary.

    Notes
    -----
    1. This routine requires PTDF matrix.
    2. Nodal price ``pi`` is calculated with three parts.
    3. Bus angle ``aBus`` is calculated after solving the problem.
    """

    def __init__(self, system, config):
        DCOPF.__init__(self, system, config)
        self.info = 'DCOPF using PTDF'
        self.type = 'DCED'

        # NOTE: in this way, we still follow the implementation that devices
        # connectivity status is considered in connection matrix
        self.ued = NumOp(u=self.Cl,
                         name='ued', tex_name=r'u_{e,d}',
                         info='Effective load connection status',
                         fun=np.sum, args=dict(axis=0),
                         no_parse=True)
        self.uesh = NumOp(u=self.Csh,
                          name='uesh', tex_name=r'u_{e,sh}',
                          info='Effective shunt connection status',
                          fun=np.sum, args=dict(axis=0),
                          no_parse=True)

        self.PTDF = RParam(info='PTDF',
                           name='PTDF', tex_name=r'P_{TDF}',
                           model='mats', src='PTDF',
                           no_parse=True, sparse=True)

        # --- rewrite Expression plf: line flow---
        self.plf.e_str = 'PTDF @ (Cg@pg - Cl@pd - Csh@gsh - Pbusinj)'

        # --- rewrite nodal price ---
        self.Cft = RParam(info='Line connection matrix',
                          name='Cft', tex_name=r'C_{ft}',
                          model='mats', src='Cft',
                          no_parse=True, sparse=True,)
        self.pilb = ExpressionCalc(info='Congestion price, dual of <plflb>',
                                   name='pilb',
                                   model='Line', src=None,
                                   e_str='plflb.dual_variables[0]')
        self.piub = ExpressionCalc(info='Congestion price, dual of <plfub>',
                                   name='piub',
                                   model='Line', src=None,
                                   e_str='plfub.dual_variables[0]')
        self.pib = ExpressionCalc(info='Energy price, dual of <pb>',
                                  name='pib',
                                  model='Bus', src=None,
                                  e_str='pb.dual_variables[0]')
        pi = 'pb.dual_variables[0] + Cft@(plfub.dual_variables[0] - plflb.dual_variables[0])'
        self.pi.e_str = pi

    def _post_solve(self):
        """Calculate aBus"""
        super()._post_solve()
        sys = self.system
        Pbus = sys.mats.Cg._v @ self.pg.v
        Pbus -= sys.mats.Cl._v @ self.pd.v
        Pbus -= sys.mats.Csh._v @ self.gsh.v
        Pbus -= self.Pbusinj.v
        aBus = sps.linalg.spsolve(sys.mats.Bbus._v, Pbus)
        slack0_uid = sys.Bus.idx2uid(sys.Slack.bus.v[0])
        self.aBus.v = aBus - aBus[slack0_uid]
        return super()._post_solve()

    def init(self, **kwargs):
        if self.system.mats.PTDF._v is None:
            logger.warning('PTDF is not available, build it now')
            self.system.mats.build_ptdf()
        return super().init(**kwargs)



"""
Module for optimization Expression.
"""
import logging

from typing import Optional
import re

import numpy as np

import cvxpy as cp

from ams.utils import pretty_long_message
from ams.shared import _prefix, _max_length

from ams.opt import OptzBase, ensure_symbols, ensure_mats_and_parsed

logger = logging.getLogger(__name__)


class Expression(OptzBase):
    """
    Base class for expressions used in a routine.

    Parameters
    ----------
    name : str, optional
        Expression name. One should typically assigning the name directly because
        it will be automatically assigned by the model. The value of ``name``
        will be the symbol name to be used in expressions.
    tex_name : str, optional
        LaTeX-formatted variable symbol. Defaults to the value of ``name``.
    info : str, optional
        Descriptive information
    unit : str, optional
        Unit
    e_str : str, optional
        Expression string
    model : str, optional
        Name of the owner model or group.
    src : str, optional
        Source expression name
    vtype : type, optional
        Value type
    horizon : ams.routines.RParam, optional
        Horizon
    """

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 e_str: Optional[str] = None,
                 model: Optional[str] = None,
                 src: Optional[str] = None,
                 vtype: Optional[str] = float,
                 horizon: Optional[str] = None,
                 ):
        OptzBase.__init__(self, name=name, info=info, unit=unit, model=model)
        self.tex_name = tex_name
        self.e_str = e_str
        self.optz = None
        self.code = None
        self.src = src
        self.horizon = horizon

    @ensure_symbols
    def parse(self):
        """
        Parse the expression.

        Returns
        -------
        bool
            Returns True if the parsing is successful, False otherwise.
        """
        sub_map = self.om.rtn.syms.sub_map
        code_expr = self.e_str
        for pattern, replacement in sub_map.items():
            try:
                code_expr = re.sub(pattern, replacement, code_expr)
            except Exception as e:
                raise Exception(f"Error in parsing expr <{self.name}>.\n{e}")
        self.code = code_expr
        msg = f" - Expression <{self.name}>: {self.e_str}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        return True

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the expression.

        Returns
        -------
        bool
            Returns True if the evaluation is successful, False otherwise.
        """
        msg = f" - Expression <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        try:
            local_vars = {'self': self, 'np': np, 'cp': cp, 'sub_map': self.om.rtn.syms.val_map}
            self.optz = self._evaluate_expression(self.code, local_vars=local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating Expression <{self.name}>.\n{e}")
        return True

    def _evaluate_expression(self, code, local_vars=None):
        """
        Helper method to evaluate the expression code.

        Parameters
        ----------
        code : str
            The code string representing the expression.

        Returns
        -------
        cp.Expression
            The evaluated cvxpy expression.
        """
        return eval(code, {}, local_vars)

    @property
    def v(self):
        """
        Return the CVXPY expression value.
        """
        if self.optz is None:
            return None
        else:
            return self.optz.value

    @v.setter
    def v(self, value):
        """
        Set the value.
        """
        raise NotImplementedError('Cannot set value to an Expression.')

    @property
    def shape(self):
        """
        Return the shape.
        """
        try:
            return self.om.__dict__[self.name].shape
        except KeyError:
            logger.warning('Shape info is not ready before initialization.')
            return None

    @property
    def size(self):
        """
        Return the size.
        """
        if self.rtn.initialized:
            return self.om.__dict__[self.name].size
        else:
            logger.warning(f'Routine <{self.rtn.class_name}> is not initialized yet.')
            return None

    @property
    def e(self):
        """
        Return the calculated expression value.
        """
        if self.code is None:
            logger.info(f"Expression <{self.name}> is not parsed yet.")
            return None

        val_map = self.om.rtn.syms.val_map
        code = self.code
        for pattern, replacement in val_map.items():
            try:
                code = re.sub(pattern, replacement, code)
            except TypeError as e:
                raise TypeError(e)

        try:
            logger.debug(pretty_long_message(f"Value code: {code}",
                                             _prefix, max_length=_max_length))
            local_vars = {'self': self, 'np': np, 'cp': cp, 'val_map': val_map}
            return self._evaluate_expression(code, local_vars)
        except Exception as e:
            logger.error(f"Error in calculating expr <{self.name}>.\n{e}")
            return None



"""
Shared constants and delayed imports.

This module is supplementary to the ``andes.shared`` module.
"""
import logging
import unittest
from functools import wraps
from time import strftime

import cvxpy as cp

from andes.utils.lazyimport import LazyImport

from andes.system import System as adSystem


logger = logging.getLogger(__name__)

sps = LazyImport('import scipy.sparse as sps')
np = LazyImport('import numpy as np')
pd = LazyImport('import pandas as pd')
ppoption = LazyImport('from pypower.ppoption import ppoption')
runpf = LazyImport('from pypower.runpf import runpf')
runopf = LazyImport('from pypower.runopf import runopf')
opf = LazyImport('from gurobi_optimods import opf')

# --- an empty ANDES system ---
empty_adsys = adSystem(autogen_stale=False)
ad_models = list(empty_adsys.models.keys())
ad_pf_models = [mname for mname, model in empty_adsys.models.items() if model.flags.pflow]
ad_dyn_models = [mname for mname, model in empty_adsys.models.items() if model.flags.tds and not model.flags.pflow]

# --- NumPy constants ---
# NOTE: In NumPy 2.0, np.Inf and np.NaN are deprecated.
inf = np.inf
nan = np.nan

# --- misc constants ---
_prefix = r" - --------------> | "  # NOQA
_max_length = 80                    # NOQA

# NOTE: copyright
copyright_msg = "Copyright (C) 2023-2025 Jinning Wang"
nowarranty_msg = "AMS comes with ABSOLUTELY NO WARRANTY"
report_time = strftime("%m/%d/%Y %I:%M:%S %p")

# NOTE: copied from CVXPY documentation, last checked on 2024/10/30, v1.5
mip_solvers = ['CBC', 'COPT', 'GLPK_MI', 'CPLEX', 'GUROBI',
               'MOSEK', 'SCIP', 'XPRESS', 'SCIPY']

misocp_solvers = ['MOSEK', 'CPLEX', 'GUROBI', 'XPRESS', 'SCIP']

installed_solvers = cp.installed_solvers()

installed_mip_solvers = [s for s in installed_solvers if s in mip_solvers]


def require_MISOCP_solver(f):
    """
    Decorator for functions that require MISOCP solver.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if not any(s in misocp_solvers for s in installed_solvers):
            raise ModuleNotFoundError("No MISOCP solver is available.")
        return f(*args, **kwargs)

    return wrapper


def require_MIP_solver(f):
    """
    Decorator for functions that require MIP solver.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if not any(s in mip_solvers for s in installed_solvers):
            raise ModuleNotFoundError("No MIP solver is available.")
        return f(*args, **kwargs)

    return wrapper


def skip_unittest_without_MIP(f):
    """
    Decorator for skipping tests that require MIP solver.
    """
    def wrapper(*args, **kwargs):
        if any(s in mip_solvers for s in installed_solvers):
            pass
        else:
            raise unittest.SkipTest("No MIP solver is available.")
        return f(*args, **kwargs)
    return wrapper


def skip_unittest_without_MISOCP(f):
    """
    Decorator for skipping tests that require MISOCP solver.
    """
    def wrapper(*args, **kwargs):
        if any(s in misocp_solvers for s in installed_solvers):
            pass
        else:
            raise unittest.SkipTest("No MISOCP solver is available.")
        return f(*args, **kwargs)
    return wrapper


def skip_unittest_without_PYPOWER(f):
    """
    Decorator for skipping tests that require PYPOWER.
    """
    def wrapper(*args, **kwargs):
        try:
            import pypower  # NOQA
        except ImportError:
            raise unittest.SkipTest("PYPOWER is not installed.")
        return f(*args, **kwargs)
    return wrapper


def skip_unittest_without_gurobi_optimods(f):
    """
    Decorator for skipping tests that require gurobi_optimods.
    """
    def wrapper(*args, **kwargs):
        try:
            import gurobi_optimods  # NOQA
        except ImportError:
            raise unittest.SkipTest("Gurobi is not installed.")
        return f(*args, **kwargs)
    return wrapper



import unittest

import ams


class TestCodegen(unittest.TestCase):
    """
    Test code generation.
    """

    def test_1_docs(self) -> None:
        sp = ams.system.System()
        out = ''
        for tp in sp.types.values():
            out += tp.doc_all()

        out = ''
        for group in sp.groups.values():
            out += group.doc_all()

    def test_docum(self) -> None:
        sp = ams.load(ams.get_case('5bus/pjm5bus_demo.json'),
                      setup=True, no_output=True)
        sp.DCOPF.init()
        docum = sp.DCOPF.docum
        for export in ['plain', 'rest']:
            docum._obj_doc(max_width=78, export=export)
            docum._constr_doc(max_width=78, export=export)
            docum._exprc_doc(max_width=78, export=export)
            docum._var_doc(max_width=78, export=export)
            docum._service_doc(max_width=78, export=export)
            docum._param_doc(max_width=78, export=export)
            docum.config.doc(max_width=78, export=export)


class TestParamCorrection(unittest.TestCase):
    """
    Test parameter correction.
    """

    def setUp(self) -> None:
        """
        Test setup.
        """
        self.ss = ams.load(ams.get_case('matpower/case14.m'),
                           setup=False, no_output=True, default_config=True,)

    def test_line_correction(self):
        """
        Test line correction.
        """
        self.ss.Line.rate_a.v[5] = 0.0
        self.ss.Line.rate_b.v[6] = 0.0
        self.ss.Line.rate_c.v[7] = 0.0
        self.ss.Line.amax.v[8] = 0.0
        self.ss.Line.amin.v[9] = 0.0

        self.ss.setup()

        self.assertIsNot(self.ss.Line.rate_a.v[5], 0.0)
        self.assertIsNot(self.ss.Line.rate_b.v[6], 0.0)
        self.assertIsNot(self.ss.Line.rate_c.v[7], 0.0)
        self.assertIsNot(self.ss.Line.amax.v[8], 0.0)
        self.assertIsNot(self.ss.Line.amin.v[9], 0.0)



"""
MATPOWER parser.
"""
import logging
import re
import numpy as np

from andes.io import read_file_like
from andes.io.xlsx import confirm_overwrite
from andes.shared import deg2rad, rad2deg

from ams import __version__ as version
from ams.shared import copyright_msg, nowarranty_msg, report_time

logger = logging.getLogger(__name__)


def testlines(infile):
    """
    Test if this file is in the MATPOWER format.

    NOT YET IMPLEMENTED.
    """

    return True  # hard coded


def read(system, file):
    """
    Read a MATPOWER data file into mpc, and build AMS device elements.
    """

    mpc = m2mpc(file)
    return mpc2system(mpc, system)


def m2mpc(infile: str) -> dict:
    """
    Parse a MATPOWER file and return a dictionary containing the parsed data.

    This function processes MATPOWER case files and extracts relevant fields
    into a structured dictionary. It is revised from ``andes.io.matpower.m2mpc``.

    Supported fields include:
    - `baseMVA`: The system base power in MVA.
    - `bus`: Bus data, including voltage, load, and generation information.
    - `bus_name`: Names of the buses (if available).
    - `gen`: Generator data, including power limits and voltage setpoints.
    - `branch`: Branch data, including line impedances and ratings.
    - `gencost`: Generator cost data (parsed but not used in this implementation).
    - `areas`: Area data (parsed but not used in this implementation).
    - `gentype`: Generator type information (if available).
    - `genfuel`: Generator fuel type information (if available).

    Parameters
    ----------
    infile : str
        Path to the MATPOWER file to be parsed.

    Returns
    -------
    dict
        A dictionary containing the parsed MATPOWER data, where keys correspond
        to MATPOWER struct names and values are numpy arrays or lists.
    """

    func = re.compile(r'function\s')
    mva = re.compile(r'\s*mpc.baseMVA\s*=\s*')
    bus = re.compile(r'\s*mpc.bus\s*=\s*\[?')
    gen = re.compile(r'\s*mpc.gen\s*=\s*\[')
    branch = re.compile(r'\s*mpc.branch\s*=\s*\[')
    area = re.compile(r'\s*mpc.areas\s*=\s*\[')
    gencost = re.compile(r'\s*mpc.gencost\s*=\s*\[')
    bus_name = re.compile(r'\s*mpc.bus_name\s*=\s*{')
    gentype = re.compile(r'\s*mpc.gentype\s*=\s*{')
    genfuel = re.compile(r'\s*mpc.genfuel\s*=\s*{')
    end = re.compile(r'\s*[\];}]')
    has_digit = re.compile(r'.*\d+\s*]?;?')

    field = None
    info = True

    mpc = {
        'version': 2,  # not in use
        'baseMVA': 100,
        'bus': [],
        'gen': [],
        'branch': [],
        'area': [],
        'gencost': [],
        'bus_name': [],
        'gentype': [],
        'genfuel': [],
    }

    input_list = read_file_like(infile)

    for line in input_list:
        line = line.strip().rstrip(';')
        if not line:
            continue
        elif func.search(line):  # skip function declaration
            continue
        elif len(line.split('%')[0]) == 0:
            if info is True:
                logger.info(line[1:])
                info = False
            else:
                continue
        elif mva.search(line):
            mpc["baseMVA"] = float(line.split('=')[1])

        if not field:
            if bus.search(line):
                field = 'bus'
            elif gen.search(line):
                field = 'gen'
            elif branch.search(line):
                field = 'branch'
            elif area.search(line):
                field = 'area'
            elif gencost.search(line):
                field = 'gencost'
            elif bus_name.search(line):
                field = 'bus_name'
            elif gentype.search(line):
                field = 'gentype'
            elif genfuel.search(line):
                field = 'genfuel'
            else:
                continue
        elif end.search(line):
            field = None
            continue

        # parse mpc sections
        if field:
            if line.find('=') >= 0:
                line = line.split('=')[1]
            if line.find('[') >= 0:
                line = re.sub(r'\[', '', line)
            elif line.find('{') >= 0:
                line = re.sub(r'{', '', line)

            if field in ['bus_name', 'gentype', 'genfuel']:
                # Handle string-based fields
                line = line.split(';')
                data = [i.strip('\'').strip() for i in line if i.strip()]
                mpc[field].extend(data)
            else:
                if not has_digit.search(line):
                    continue
                line = line.split('%')[0].strip()
                line = line.split(';')
                for item in line:
                    if not has_digit.search(item):
                        continue
                    try:
                        data = np.array([float(val) for val in item.split()])
                    except Exception as e:
                        logger.error('Error parsing "%s"', infile)
                        raise e
                    mpc[field].append(data)

    # convert mpc to np array
    mpc_array = dict()
    for key, val in mpc.items():
        if isinstance(val, (float, int)):
            mpc_array[key] = val
        elif isinstance(val, list):
            if len(val) == 0:
                continue
            if key in ['bus_name', 'gentype', 'genfuel']:
                mpc_array[key] = np.array(val, dtype=object)
            else:
                mpc_array[key] = np.array(val)
        else:
            raise NotImplementedError("Unknown type for mpc, ", type(val))
    return mpc_array


def mpc2system(mpc: dict, system) -> bool:
    """
    Load an mpc dict into an empty AMS system.

    Revised from ``andes.io.matpower.mpc2system``.

    Note that `mbase` in mpc is converted to `Sn`, but it is not actually used in
    MATPOWER nor AMS.

    In converted AMS system, StaticGen idxes are 1-based, while the sequence follow
    the order of the original MATPOWER data.

    Parameters
    ----------
    system : ams.system.System
        Empty system to load the data into.
    mpc : dict
        mpc struct names : numpy arrays

    Returns
    -------
    bool
        True if successful, False otherwise.
    """

    # list of buses with slack gen
    sw = []

    system.config.mva = base_mva = mpc['baseMVA']

    for data in mpc['bus']:
        # idx  ty   pd   qd  gs  bs  area  vmag  vang  baseKV  zone  vmax  vmin
        # 0    1    2    3   4   5   6     7     8     9       10    11    12
        idx = int(data[0])
        ty = data[1]
        if ty == 3:
            sw.append(idx)
        pd = data[2] / base_mva
        qd = data[3] / base_mva
        gs = data[4] / base_mva
        bs = data[5] / base_mva
        area = data[6]
        vmag = data[7]
        vang = data[8] * deg2rad
        baseKV = data[9]
        if baseKV == 0:
            baseKV = 110
        zone = data[10]
        vmax = data[11]
        vmin = data[12]

        system.add('Bus', idx=idx, name=None,
                   type=ty, Vn=baseKV,
                   v0=vmag, a0=vang,
                   vmax=vmax, vmin=vmin,
                   area=area, zone=zone)
        if pd != 0 or qd != 0:
            system.add('PQ', bus=idx, name=None, Vn=baseKV, p0=pd, q0=qd)
        if gs or bs:
            system.add('Shunt', bus=idx, name=None, Vn=baseKV, g=gs, b=bs)

    gen_idx = 0
    if mpc['gen'].shape[1] <= 10:  # missing data
        mpc_gen = np.zeros((mpc['gen'].shape[0], 21), dtype=np.float64)
        mpc_gen[:, :10] = mpc['gen']
        mpc_gen[:, 16] = system.PV.Ragc.default * base_mva / 60
        mpc_gen[:, 17] = system.PV.R10.default * base_mva
        mpc_gen[:, 18] = system.PV.R30.default * base_mva
        mpc_gen[:, 19] = system.PV.Rq.default * base_mva / 60
    else:
        mpc_gen = mpc['gen']

    # Ensure 'gentype' and 'genfuel' keys exist in mpc, with default values if missing
    gentype = mpc.get('gentype', [''] * mpc_gen.shape[0])
    genfuel = mpc.get('genfuel', [''] * mpc_gen.shape[0])

    # Validate lengths of 'gentype' and 'genfuel' against the number of generators
    if len(gentype) != mpc_gen.shape[0]:
        raise ValueError(
            f"'gentype' length ({len(gentype)}) does not match the number of generators ({mpc_gen.shape[0]})")
    if len(genfuel) != mpc_gen.shape[0]:
        raise ValueError(
            f"'genfuel' length ({len(genfuel)}) does not match the number of generators ({mpc_gen.shape[0]})")

    for data, gt, gf in zip(mpc_gen, gentype, genfuel):
        # bus  pg  qg  qmax  qmin  vg  mbase  status  pmax  pmin
        # 0    1   2   3     4     5   6      7       8     9
        # pc1  pc2  qc1min  qc1max  qc2min  qc2max  ramp_agc  ramp_10
        # 10   11   12      13      14      15      16        17
        # ramp_30  ramp_q  apf
        # 18       19      20
        bus_idx = int(data[0])
        gen_idx += 1
        vg = data[5]
        status = int(data[7])
        pg = data[1] / base_mva
        qg = data[2] / base_mva
        qmax = data[3] / base_mva
        qmin = data[4] / base_mva
        pmax = data[8] / base_mva
        pmin = data[9] / base_mva
        pc1 = data[10] / base_mva
        pc2 = data[11] / base_mva
        qc1min = data[12] / base_mva
        qc1max = data[13] / base_mva
        qc2min = data[14] / base_mva
        qc2max = data[15] / base_mva
        ramp_agc = 60 * data[16] / base_mva  # MW/min -> MW/h
        ramp_10 = data[17] / base_mva  # MW -> MW/h
        ramp_30 = data[18] / base_mva  # MW -> MW/h
        ramp_q = 60 * data[19] / base_mva  # MVAr/min -> MVAr/h
        apf = data[20]

        uid = system.Bus.idx2uid(bus_idx)
        vn = system.Bus.Vn.v[uid]
        a0 = system.Bus.a0.v[uid]

        if bus_idx in sw:
            system.add('Slack', idx=gen_idx, bus=bus_idx, busr=bus_idx,
                       name=None,
                       u=status, Sn=data[6],
                       Vn=vn, v0=vg, p0=pg, q0=qg, a0=a0,
                       pmax=pmax, pmin=pmin,
                       qmax=qmax, qmin=qmin,
                       Pc1=pc1, Pc2=pc2,
                       Qc1min=qc1min, Qc1max=qc1max,
                       Qc2min=qc2min, Qc2max=qc2max,
                       Ragc=ramp_agc, R10=ramp_10,
                       R30=ramp_30, Rq=ramp_q,
                       apf=apf, gentype=gt, genfuel=gf)
        else:
            system.add('PV', idx=gen_idx, bus=bus_idx, busr=bus_idx,
                       name=None,
                       u=status, Sn=data[6],
                       Vn=vn, v0=vg, p0=pg, q0=qg,
                       pmax=pmax, pmin=pmin,
                       qmax=qmax, qmin=qmin,
                       Pc1=pc1, Pc2=pc2,
                       Qc1min=qc1min, Qc1max=qc1max,
                       Qc2min=qc2min, Qc2max=qc2max,
                       Ragc=ramp_agc, R10=ramp_10,
                       R30=ramp_30, Rq=ramp_q,
                       apf=apf, gentype=gt, genfuel=gf)

    for data in mpc['branch']:
        # fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle
        # 0     1       2   3   4   5       6       7       8       9
        # status	angmin	angmax	Pf	Qf	Pt	Qt
        # 10        11      12      13  14  15  16
        fbus = int(data[0])
        tbus = int(data[1])
        r = data[2]
        x = data[3]
        b = data[4]
        rate_a = data[5] / base_mva
        rate_b = data[6] / base_mva
        rate_c = data[7] / base_mva
        amin = data[11] * deg2rad
        amax = data[12] * deg2rad

        status = int(data[10])

        if (data[8] == 0.0) or (data[8] == 1.0 and data[9] == 0.0):
            # not a transformer
            tf = False
            ratio = 1
            angle = 0
        else:
            tf = True
            ratio = data[8]
            angle = data[9] * deg2rad

        vf = system.Bus.Vn.v[system.Bus.idx2uid(fbus)]
        vt = system.Bus.Vn.v[system.Bus.idx2uid(tbus)]
        system.add('Line', u=status, name=f'Line {fbus:.0f}-{tbus:.0f}',
                   Vn1=vf, Vn2=vt,
                   bus1=fbus, bus2=tbus,
                   r=r, x=x, b=b,
                   trans=tf, tap=ratio, phi=angle,
                   rate_a=rate_a, rate_b=rate_b, rate_c=rate_c,
                   amin=amin, amax=amax)

    if ('bus_name' in mpc) and (len(mpc['bus_name']) == len(system.Bus.name.v)):
        system.Bus.name.v[:] = mpc['bus_name']

    # --- gencost ---
    if 'gencost' in mpc:
        gcost_idx = 0
        gen_idx = np.arange(mpc['gen'].shape[0]) + 1
        mpc_cost = mpc['gencost']
        if mpc_cost[0, 0] == 1:
            logger.warning("Type 1 gencost detected, which is not supported in AMS.\n"
                           "Default type 2 cost parameters will be used as a fallback.\n"
                           "It is recommended to manually convert the gencost data to type 2.")
            mpc_cost = np.repeat(np.array([[2, 0, 0, 3, 0, 0, 0]]),
                                 mpc_cost.shape[0], axis=0)
        for data, gen in zip(mpc_cost, gen_idx):
            # NOTE: only type 2 costs are supported for now
            # type  startup shutdown	n	c2  c1  c0
            # 0     1       2           3   4   5   6
            if data[0] != 2:
                raise ValueError('Only MODEL 2 costs are supported')
            gcost_idx += 1
            gctype = int(data[0])
            startup = data[1]
            shutdown = data[2]
            if data[3] == 3:
                c2 = data[4] * base_mva ** 2
                c1 = data[5] * base_mva
                c0 = data[6]
            elif data[3] == 2:
                c2 = 0
                c1 = data[4] * base_mva
                c0 = data[5]
            else:
                raise ValueError('Unrecognized gencost model, please use eighter quadratic or linear cost model')
            system.add('GCost', gen=int(gen),
                       u=1, type=gctype,
                       idx=gcost_idx,
                       name=None,
                       csu=startup, csd=shutdown,
                       c2=c2, c1=c1, c0=c0
                       )

    # --- Area ---
    area = system.Bus.area.v
    area_map = {}
    if area:
        for a in set(area):
            a_new = system.add('Area',
                               param_dict=dict(idx=a, name=a))
            area_map[a] = a_new
        system.Bus.area.v = [area_map[a] for a in area]

    # --- Zone ---
    zone = system.Bus.zone.v
    zone_map = {}
    if zone:
        n_zone = system.Area.n
        for z in set(zone):
            z_new = system.add('Zone',
                               param_dict=dict(idx=int(n_zone + 1),
                                               name=f'{n_zone + 1}'))
            zone_map[z] = z_new
            n_zone += 1
        system.Bus.zone.v = [zone_map[z] for z in zone]

    return True


def _get_bus_id_caller(bus):
    """
    Helper function to get the bus id. Force bus id to be uid+1.

    Parameters
    ----------
    bus : andes.models.bus.Bus
        Bus object

    Returns
    -------
    lambda function to that takes bus idx and returns bus id for matpower case
    """

    if np.array(bus.idx.v).dtype in ['int', 'float']:
        return lambda x: x
    else:
        return lambda x: list(np.array(bus.idx2uid(x)) + 1)


def system2mpc(system) -> dict:
    """
    Convert a **setup** AMS system to a MATPOWER mpc dictionary.

    This function is revised from ``andes.io.matpower.system2mpc``.

    In the ``gen`` section, slack generators are listed before PV generators.

    In the converted MPC, the indices of area (bus[:, 6]) and zone (bus[:, 10])
    may differ from the original MPC. However, the mapping relationship is preserved.
    For example, if the original MPC numbers areas starting from 1, the converted
    MPC may number them starting from 0.

    The coefficients ``c2`` and ``c1`` in the generator cost data are scaled by
    ``base_mva`` to match MATPOWER's unit convention (MW).

    Parameters
    ----------
    system : ams.core.system.System
        The AMS system to be converted.

    Returns
    -------
    mpc : dict
        A dictionary in MATPOWER format representing the converted AMS system.
    """

    mpc = dict(version='2',
               baseMVA=system.config.mva,
               bus=np.zeros((system.Bus.n, 13), dtype=np.float64),
               gen=np.zeros((system.PV.n + system.Slack.n, 21), dtype=np.float64),
               branch=np.zeros((system.Line.n, 17), dtype=np.float64),
               gencost=np.zeros((system.GCost.n, 7), dtype=np.float64),
               )

    if not system.is_setup:
        logger.warning("System is not setup and will be setup now.")
        system.setup()

    if system.Bus.name.v is not None:
        mpc['bus_name'] = system.Bus.name.v

    base_mva = system.config.mva

    # --- Bus ---
    bus = mpc['bus']
    gen = mpc['gen']

    to_busid = _get_bus_id_caller(system.Bus)

    bus[:, 0] = to_busid(system.Bus.idx.v)
    bus[:, 1] = 1
    if system.Area.n > 0:
        bus[:, 6] = system.Area.idx2uid(system.Bus.area.v)
    bus[:, 7] = system.Bus.v0.v
    bus[:, 8] = system.Bus.a0.v * rad2deg
    bus[:, 9] = system.Bus.Vn.v
    if system.Zone.n > 0:
        bus[:, 10] = system.Zone.idx2uid(system.Bus.zone.v)
    bus[:, 11] = system.Bus.vmax.v
    bus[:, 12] = system.Bus.vmin.v

    # --- PQ ---
    if system.PQ.n > 0:
        pq_pos = system.Bus.idx2uid(system.PQ.bus.v)
        u = system.PQ.u.v
        bus[pq_pos, 2] = u * system.PQ.p0.v * base_mva
        bus[pq_pos, 3] = u * system.PQ.q0.v * base_mva

    # --- Shunt ---
    if system.Shunt.n > 0:
        shunt_pos = system.Bus.idx2uid(system.Shunt.bus.v)
        bus[shunt_pos, 4] = system.Shunt.g.v * base_mva
        bus[shunt_pos, 5] = system.Shunt.b.v * base_mva

    # --- PV ---
    if system.PV.n > 0:
        PV = system.PV
        pv_pos = system.Bus.idx2uid(PV.bus.v)
        bus[pv_pos, 1] = 2
        gen[system.Slack.n:, 0] = to_busid(PV.bus.v)
        gen[system.Slack.n:, 1] = PV.p0.v * base_mva
        gen[system.Slack.n:, 2] = PV.q0.v * base_mva
        gen[system.Slack.n:, 3] = PV.qmax.v * base_mva
        gen[system.Slack.n:, 4] = PV.qmin.v * base_mva
        gen[system.Slack.n:, 5] = PV.v0.v
        gen[system.Slack.n:, 6] = PV.Sn.v
        gen[system.Slack.n:, 7] = PV.u.v
        gen[system.Slack.n:, 8] = (PV.ctrl.v * PV.pmax.v + (1 - PV.ctrl.v) * PV.p0.v) * base_mva
        gen[system.Slack.n:, 9] = (PV.ctrl.v * PV.pmin.v + (1 - PV.ctrl.v) * PV.p0.v) * base_mva
        gen[system.Slack.n:, 16] = PV.Ragc.v * base_mva * 60    # MW/h -> MW/min
        gen[system.Slack.n:, 17] = PV.R10.v * base_mva
        gen[system.Slack.n:, 18] = PV.R30.v * base_mva
        gen[system.Slack.n:, 19] = PV.Rq.v * base_mva * 60  # MVAr/h -> MVAr/min

    # --- Slack ---
    if system.Slack.n > 0:
        slack_pos = system.Bus.idx2uid(system.Slack.bus.v)
        bus[slack_pos, 1] = 3
        bus[slack_pos, 8] = system.Slack.a0.v * rad2deg
        gen[:system.Slack.n, 0] = to_busid(system.Slack.bus.v)
        gen[:system.Slack.n, 1] = system.Slack.p0.v * base_mva
        gen[:system.Slack.n, 2] = system.Slack.q0.v * base_mva
        gen[:system.Slack.n, 3] = system.Slack.qmax.v * base_mva
        gen[:system.Slack.n, 4] = system.Slack.qmin.v * base_mva
        gen[:system.Slack.n, 5] = system.Slack.v0.v
        gen[:system.Slack.n, 6] = system.Slack.Sn.v
        gen[:system.Slack.n, 7] = system.Slack.u.v
        gen[:system.Slack.n, 8] = system.Slack.pmax.v * base_mva
        gen[:system.Slack.n, 9] = system.Slack.pmin.v * base_mva
        gen[:system.Slack.n, 16] = system.Slack.Ragc.v * base_mva * 60    # MW/h -> MW/min
        gen[:system.Slack.n, 17] = system.Slack.R10.v * base_mva
        gen[:system.Slack.n, 18] = system.Slack.R30.v * base_mva
        gen[:system.Slack.n, 19] = system.Slack.Rq.v * base_mva * 60    # MVAr/h -> MVAr/min

    if system.Line.n > 0:
        branch = mpc['branch']
        branch[:, 0] = to_busid(system.Line.bus1.v)
        branch[:, 1] = to_busid(system.Line.bus2.v)
        branch[:, 2] = system.Line.r.v
        branch[:, 3] = system.Line.x.v
        branch[:, 4] = system.Line.b.v
        branch[:, 5] = system.Line.rate_a.v * base_mva
        branch[:, 6] = system.Line.rate_b.v * base_mva
        branch[:, 7] = system.Line.rate_c.v * base_mva
        branch[:, 8] = system.Line.tap.v
        branch[:, 9] = system.Line.phi.v * rad2deg
        branch[:, 10] = system.Line.u.v
        branch[:, 11] = system.Line.amin.v * rad2deg
        branch[:, 12] = system.Line.amax.v * rad2deg

    # --- GCost ---
    # NOTE: adjust GCost sequence to match the generator sequence
    if system.GCost.n > 0:
        stg_idx = system.Slack.idx.v + system.PV.idx.v
        gcost_idx = system.GCost.find_idx(keys=['gen'], values=[stg_idx])
        gcost_uid = system.GCost.idx2uid(gcost_idx)
        gencost = mpc['gencost']
        gencost[:, 0] = system.GCost.type.v[gcost_uid]
        gencost[:, 1] = system.GCost.csu.v[gcost_uid]
        gencost[:, 2] = system.GCost.csd.v[gcost_uid]
        gencost[:, 3] = 3
        gencost[:, 4] = system.GCost.c2.v[gcost_uid] / base_mva / base_mva
        gencost[:, 5] = system.GCost.c1.v[gcost_uid] / base_mva
        gencost[:, 6] = system.GCost.c0.v[gcost_uid]
    else:
        mpc.pop('gencost')

    # --- gentype ---
    stg = system.StaticGen.get_all_idxes()
    gentype = system.StaticGen.get(src='gentype', attr='v', idx=stg)
    if any(gentype):
        mpc['gentype'] = np.array(gentype)

    # --- genfuel ---
    genfuel = system.StaticGen.get(src='genfuel', attr='v', idx=stg)
    if any(genfuel):
        mpc['genfuel'] = np.array(genfuel)

    # --- Bus Name ---
    if any(system.Bus.name.v):
        mpc['bus_name'] = np.array(system.Bus.name.v)

    return mpc


def mpc2m(mpc: dict, outfile: str) -> str:
    """
    Write a MATPOWER mpc dict to a M-file.

    Parameters
    ----------
    mpc : dict
        MATPOWER mpc dictionary.
    outfile : str
        Path to the output M-file.
    """
    with open(outfile, 'w') as f:
        # Add version info
        f.write(f"%% Converted by AMS {version}\n")
        f.write(f"%% {copyright_msg}\n\n")
        f.write(f"%% {nowarranty_msg}\n")
        f.write(f"%% Convert time: {report_time}\n\n")

        f.write("function mpc = mpc_case\n")
        f.write("mpc.version = '2';\n\n")

        # Write baseMVA
        f.write(f"%% system MVA base\nmpc.baseMVA = {mpc['baseMVA']};\n\n")

        # Write bus data
        f.write("%% bus data\n")
        f.write("%% bus_i type Pd Qd Gs Bs area Vm Va baseKV zone Vmax Vmin\n")
        f.write("mpc.bus = [\n")
        for row in mpc['bus']:
            f.write("    " + "\t".join(f"{val:.6g}" for val in row) + ";\n")
        f.write("];\n\n")

        # Write generator data
        f.write("%% generator data\n")
        f.write("%% bus Pg Qg Qmax Qmin Vg mBase status Pmax Pmin\n")
        f.write("%% Pc1 Pc2 Qc1min Qc1max Qc2min Qc2max ramp_agc ramp_10 ramp_30 ramp_q apf\n")
        f.write("mpc.gen = [\n")
        for row in mpc['gen']:
            f.write("    " + "\t".join(f"{val:.6g}" for val in row) + ";\n")
        f.write("];\n\n")

        # Write branch data
        f.write("%% branch data\n")
        f.write("%% fbus tbus r x b rateA rateB rateC ratio angle status angmin angmax PF QF PT QT\n")
        f.write("mpc.branch = [\n")
        for row in mpc['branch']:
            f.write("    " + "\t".join(f"{val:.6g}" for val in row) + ";\n")
        f.write("];\n\n")

        # Write generator cost data if available
        if 'gencost' in mpc:
            f.write("%% generator cost data\n")
            f.write("%% 1 startup shutdown n x1 y1 ... xn yn\n")
            f.write("%% 2 startup shutdown n c(n-1) ... c0\n")
            f.write("mpc.gencost = [\n")
            for row in mpc['gencost']:
                f.write("    " + "\t".join(f"{val:.6g}" for val in row) + ";\n")
            f.write("];\n\n")

        # Write bus names if available and not all None
        if 'bus_name' in mpc and any(mpc['bus_name']):
            f.write("%% bus names\n")
            f.write("mpc.bus_name = {\n")
            for name in mpc['bus_name']:
                f.write(f"    '{name}';\n")
            f.write("};\n\n")

        # Write generator types if available and not all None
        if 'gentype' in mpc and any(mpc['gentype']):
            f.write("%% generator types\n")
            f.write("mpc.gentype = {\n")
            for gentype in mpc['gentype']:
                f.write(f"    '{gentype}';\n")
            f.write("};\n\n")

        # Write generator fuels if available and not all None
        if 'genfuel' in mpc and any(mpc['genfuel']):
            f.write("%% generator fuels\n")
            f.write("mpc.genfuel = {\n")
            for genfuel in mpc['genfuel']:
                f.write(f"    '{genfuel}';\n")
            f.write("};\n\n")

    logger.info(f"Finished writing MATPOWER case to {outfile}")
    return outfile


def write(system, outfile: str, overwrite: bool = None) -> bool:
    """
    Export an AMS system to a MATPOWER M-file.

    This function converts an AMS system object into a MATPOWER-compatible
    mpc dictionary and writes it to a specified output file in MATPOWER format.

    In the converted MPC, the indices of area (bus[:, 6]) and zone (bus[:, 10])
    may differ from the original MPC. However, the mapping relationship is preserved.
    For example, if the original MPC numbers areas starting from 1, the converted
    MPC may number them starting from 0.

    Parameters
    ----------
    system : ams.system.System
        A loaded system.
    outfile : str
        Path to the output file.
    overwrite : bool, optional
        None to prompt for overwrite selection; True to overwrite; False to not overwrite.

    Returns
    -------
    bool
        True if the file was successfully written, False otherwise.
    """
    if not confirm_overwrite(outfile, overwrite=overwrite):
        return False

    mpc = system2mpc(system)
    mpc2m(mpc, outfile)
    logger.info('MATPOWER m case file written to "%s"', outfile)
    return True



import unittest
import os

import ams.main
import ams.cli


class TestCLI(unittest.TestCase):

    def test_cli_parser(self):
        ams.cli.create_parser()

    def test_cli_preamble(self):
        ams.cli.preamble()

    def test_main_doc(self):
        ams.main.doc('Bus')
        ams.main.doc(list_supported=True)

    def test_versioninfo(self):
        ams.main.versioninfo()

    def test_misc(self):
        ams.main.misc(show_license=True)
        ams.main.misc(save_config=None, overwrite=True)

    def test_profile_run(self):
        _ = ams.main.run(ams.get_case('matpower/case5.m'),
                         no_output=False,
                         profile=True,)
        self.assertTrue(os.path.exists('case5_prof.prof'))
        self.assertTrue(os.path.exists('case5_prof.txt'))
        os.remove('case5_prof.prof')
        os.remove('case5_prof.txt')



import unittest
import numpy as np

import ams
from ams.core.matprocessor import MParam
from ams.core.service import NumOp, NumOpDual, ZonalSum


class TestService(unittest.TestCase):
    """
    Test functionality of Services.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("ieee39/ieee39_uced_esd1.xlsx"),
                           default_config=True,
                           no_output=True,)
        self.nR = self.ss.Zone.n
        self.nB = self.ss.Bus.n
        self.nL = self.ss.Line.n
        self.nD = self.ss.StaticLoad.n  # number of static loads

    def test_NumOp_norfun(self):
        """
        Test `NumOp` without return function.
        """
        CftT = NumOp(u=self.ss.mats.Cft, fun=np.transpose)
        np.testing.assert_array_equal(CftT.v.transpose(), self.ss.mats.Cft.v)

    def test_NumOp_rfun(self):
        """
        Test `NumOp` with return function.
        """
        CftTT = NumOp(u=self.ss.mats.Cft, fun=np.transpose, rfun=np.transpose)
        np.testing.assert_array_equal(CftTT.v, self.ss.mats.Cft.v)

    def test_NumOp_ArrayOut(self):
        """
        Test `NumOp` non-array output.
        """
        M = NumOp(u=self.ss.PV.pmax,
                  fun=np.max,
                  rfun=np.dot, rargs=dict(b=10),
                  array_out=True,)
        M2 = NumOp(u=self.ss.PV.pmax,
                   fun=np.max,
                   rfun=np.dot, rargs=dict(b=10),
                   array_out=False,)
        self.assertIsInstance(M.v, np.ndarray)
        self.assertIsInstance(M2.v, (int, float))

    def test_NumOpDual(self):
        """
        Test `NumOpDual`.
        """
        p_vec = MParam(v=self.ss.DCOPF.pd.v)
        one_vec = MParam(v=np.ones(self.ss.StaticLoad.n))
        p_sum = NumOpDual(u=p_vec, u2=one_vec,
                          fun=np.multiply, rfun=np.sum)
        self.assertEqual(p_sum.v, self.ss.PQ.p0.v.sum())

    def test_ZonalSum(self):
        """
        Test `ZonalSum`.
        """
        ds = ZonalSum(u=self.ss.RTED.zd, zone="Area",
                      name="ds", tex_name=r"S_{d}",
                      info="Sum pl vector in shape of area",)
        ds.rtn = self.ss.RTED
        # check if the shape is correct
        np.testing.assert_array_equal(ds.v.shape, (self.nR, self.nD))
        # check if the values are correct
        self.assertTrue(np.all(ds.v.sum(axis=1) <= np.array([self.nB, self.nB])))



"""
The package for models used in scheduling modeling.

The file_classes includes the list of file classes and their corresponding classes.
"""


ams_file_classes = list([
    ('info', ['Summary']),
    ('bus', ['Bus']),
    ('static', ['PQ', 'Slack', 'PV']),
    ('shunt', ['Shunt']),
    ('line', ['Line', 'Jumper']),
    ('distributed', ['PVD1', 'ESD1', 'EV1', 'EV2']),
    ('renewable', ['REGCA1', 'REGCV1', 'REGCV2']),
    ('area', ['Area']),
    ('zone', ['Zone']),
    ('reserve', ['SFR', 'SR', 'NSR', 'VSGR']),
    ('cost', ['GCost', 'SFRCost', 'SRCost', 'NSRCost', 'VSGCost']),
    ('cost', ['DCost']),
    ('timeslot', ['EDTSlot', 'UCTSlot']),
])

file_classes = ams_file_classes



"""
Routines using gurobi-optimods.
"""
import logging

import io

import scipy.io

from andes.shared import np, pd
from andes.utils.misc import elapsed

from ams.io.pypower import system2ppc

from ams.opt import Var, Objective

from ams.routines.pypower import DCPF1
from ams.shared import opf

logger = logging.getLogger(__name__)


class OPF(DCPF1):
    """
    Optimal Power Flow (OPF) routine using gurobi-optimods.

    This class provides an interface for performing optimal power flow analysis
    with gurobi-optimods, supporting both AC and DC OPF formulations.

    In addition to optimizing generator dispatch, this routine can also optimize
    transmission line statuses (branch switching), enabling topology optimization.
    Refer to the gurobi-optimods documentation for further details:

    https://gurobi-optimods.readthedocs.io/en/stable/mods/opf/opf.html
    """

    def __init__(self, system, config):
        DCPF1.__init__(self, system, config)
        self.info = 'Optimal Power Flow'
        self.type = 'ACED'

        self.obj = Objective(name='obj',
                             info='total cost, placeholder',
                             e_str='sum(c2 * pg**2 + c1 * pg + c0)',
                             sense='min',)

        self.pi = Var(info='Lagrange multiplier on real power mismatch',
                      name='pi', unit='$/p.u.',
                      model='Bus', src=None,)

        self.uld = Var(info='Line commitment decision',
                       name='uld', tex_name=r'u_{l,d}',
                       model='Line', src='u',)

    def solve(self, **kwargs):
        ppc = system2ppc(self.system)
        mat = io.BytesIO()
        scipy.io.savemat(mat, {'mpc': ppc})
        mat.seek(0)
        res = opf.solve_opf(opf.read_case_matpower(mat), **kwargs)
        return res

    def unpack(self, res, **kwargs):
        """
        Unpack the results from the gurobi-optimods.
        """
        # NOTE: Map gurobi-optimods results to PPC-compatible format.
        # Only relevant columns are populated, as required by `DCOPF.unpack()`.
        # If future versions of gurobi-optimods provide additional outputs,
        # this mapping may need to be updated to extract and assign new fields.

        res_new = dict()
        res_new['success'] = res['success']
        res_new['et'] = res['et']
        res_new['f'] = res['f']
        res_new['baseMVA'] = res['baseMVA']

        bus = pd.DataFrame(res['bus'])
        res_new['bus'] = np.zeros((self.system.Bus.n, 17))
        res_new['bus'][:, 7] = bus['Vm'].values
        res_new['bus'][:, 8] = bus['Va'].values
        # NOTE: As of v2.3.2, gurobi-optimods does not return LMP

        gen = pd.DataFrame(res['gen'])
        res_new['gen'] = np.zeros((self.system.StaticGen.n, 14))
        res_new['gen'][:, 1] = gen['Pg'].values
        res_new['gen'][:, 2] = gen['Qg'].values

        branch = pd.DataFrame(res['branch'])
        res_new['branch'] = np.zeros((self.system.Line.n, 14))
        res_new['branch'][:, 13] = branch['Pf'].values
        # NOTE: unpack branch_switching decision
        res_new['branch'][:, 10] = branch['switching'].values
        return super().unpack(res_new)

    def run(self, **kwargs):
        """
        Run the OPF routine using gurobi-optimods.

        This method invokes `self.solve(**kwargs)`, which internally utilizes
        `gurobi-optimods` to solve the OPF problem.

        Keyword arguments
        -------------------
        - ``opftype`` : str
            Type of OPF to solve (default: 'AC').
        - ``branch_switching`` : bool
            Enable branch switching (default: False).
        - ``min_active_branches`` : float
            Defines the minimum number of branches that must be turned on when
            branch switching is active, i.e. the minimum number of turned on
            branches is equal to ``numbranches * min_active_branches``. Has no
            effect if ``branch_switching`` is set to False.
        - ``use_mip_start`` : bool
            Use MIP start (default: False).
        - ``time_limit`` : float
            Time limit for the solver (default: 0.0, no limit).
        """
        if not self.initialized:
            self.init()
        t0, _ = elapsed()

        # --- solve optimization ---
        t0, _ = elapsed()
        res = self.solve(**kwargs)
        self.converged = res['success']
        self.exit_code = 0 if res['success'] else 1
        _, s = elapsed(t0)
        self.exec_time = float(s.split(" ")[0])
        try:
            n_iter = res['raw']['output']['iterations']
        except Exception:
            n_iter = -1
        n_iter_str = f"{n_iter} iterations " if n_iter > 1 else f"{n_iter} iteration "
        if self.exit_code == 0:
            msg = f"<{self.class_name}> converged in {s}, "
            msg += n_iter_str + "with gurobi-optimods."
            logger.warning(msg)
            try:
                self.unpack(res)
            except Exception as e:
                logger.error(f"Failed to unpack results from {self.class_name}.\n{e}")
                return False
            self.system.report()
            return True
        else:
            msg = f"{self.class_name} failed to converge in {s}, "
            msg += n_iter_str + "with gurobi-optimods."
            logger.warning(msg)
            return False



"""
PYPOWER reader for AMS.
"""
import logging

import numpy as np

from ams.io.matpower import mpc2system, system2mpc

logger = logging.getLogger(__name__)


def testlines(infile):
    """
    Test if this file is in the PYPOWER format.

    NOT YET IMPLEMENTED.
    """
    return True


def read(system, file):
    """
    Read a PYPOWER case file into ppc and return an AMS system by calling ``ppc2system``.

    Parameters
    ----------
    system : ams.system
        Empty AMS system to load data into.
    file : str
        The path to the PYPOWER file.

    Returns
    -------
    system : ams.system.System
        The AMS system that loaded the data.
    """
    ppc = py2ppc(file)
    return ppc2system(ppc, system)


def py2ppc(infile: str) -> dict:
    """
    Parse PYPOWER file and return a dictionary with the data.

    Parameters
    ----------
    infile : str
        The path to the PYPOWER file.

    Returns
    -------
    ppc : dict
        The PYPOWER case dict.
    """
    exec(open(f"{infile}").read())
    for name, value in locals().items():
        # Check if the variable name starts with "case"
        if name.startswith("case"):
            ppc = value()
    return ppc


def ppc2system(ppc: dict, system) -> bool:
    """
    Alias for ``mpc2system``. Refer to :py:mod:`ams.io.matpower.mpc2system` for more details.

    Load an PYPOWER case dict into an empth AMS system.

    Parameters
    ----------
    ppc : dict
        The PYPOWER case dict.
    system : ams.system
        Empty AMS system to load data into.

    Returns
    -------
    bool
        True if successful; False otherwise.
    """
    return mpc2system(ppc, system)


def system2ppc(system) -> dict:
    """
    Alias for ``system2mpc``. Refer to :py:mod:`ams.io.matpower.system2mpc` for more details.

    Convert data from an AMS system to an mpc dict.

    In the ``gen`` section, slack generators preceeds PV generators.
    """
    mpc = system2mpc(system)
    np.set_printoptions(suppress=True)
    # Map the original bus indices to consecutive values
    # Adjust discontinuous bus indices
    BUS_I = mpc['bus'][:, 0].astype(int)
    bus_map = {busi0: i for i, busi0 in enumerate(BUS_I)}
    mpc['bus'][:, 0] = np.array([bus_map[busi0] for busi0 in BUS_I])
    mpc['gen'][:, 0] = np.array([bus_map[busi0] for busi0 in mpc['gen'][:, 0].astype(int)])
    mpc['branch'][:, 0] = np.array([bus_map[busi0] for busi0 in mpc['branch'][:, 0].astype(int)])
    mpc['branch'][:, 1] = np.array([bus_map[busi0] for busi0 in mpc['branch'][:, 1].astype(int)])
    return mpc


def ppc2py(ppc: dict, outfile: str) -> str:
    """
    Placeholder.
    """
    return NotImplemented


def write(system, outfile: str) -> bool:
    """
    Placeholder.
    """
    return NotImplemented



"""
Symbolic processor class for AMS routines.

This module is revised from ``andes.core.symprocessor``.
"""

import logging
from collections import OrderedDict

import sympy as sp

from andes.utils.misc import elapsed
from ams.core.matprocessor import MatProcessor

logger = logging.getLogger(__name__)


class SymProcessor:
    """
    Class for symbolic processing in AMS routine.

    Parameters
    ----------
    parent : ams.routines.base.BaseRoutine
        Routine instance to process.

    Attributes
    ----------
    sub_map : dict
        Substitution map for symbolic processing.
    tex_map : dict
        Tex substitution map for documentation.
    val_map : dict
        Value substitution map for post-solving value evaluation.
    """

    def __init__(self, parent):
        self.parent = parent
        self.inputs_dict = OrderedDict()
        self.services_dict = OrderedDict()
        self.config = parent.config
        self.class_name = parent.class_name
        self.tex_names = OrderedDict()
        self.tex_map = OrderedDict()

        lang = "cp"  # TODO: might need to be generalized to other solvers
        # only used for CVXPY
        self.sub_map = OrderedDict([
            (r'\b(\w+)\s*\*\s*(\w+)\b', r'\1 @ \2'),
            (r'\b(\w+)\s+dot\s+(\w+)\b', r'\1 * \2'),
            (r' dot ', r' * '),
            (r'\bsum\b', f'{lang}.sum'),
            (r'\bvar\b', f'{lang}.Variable'),
            (r'\bparam\b', f'{lang}.Parameter'),
            (r'\bconst\b', f'{lang}.Constant'),
            (r'\bproblem\b', f'{lang}.Problem'),
            (r'\bmultiply\b', f'{lang}.multiply'),
            (r'\bmul\b', f'{lang}.multiply'),  # alias for multiply
            (r'\bvstack\b', f'{lang}.vstack'),
            (r'\bnorm\b', f'{lang}.norm'),
            (r'\bpos\b', f'{lang}.pos'),
            (r'\bpower\b', f'{lang}.power'),
            (r'\bsign\b', f'{lang}.sign'),
            (r'\bsquare\b', f'{lang}.square'),
            (r'\bquad_over_lin\b', f'{lang}.quad_over_lin'),
            (r'\bdiag\b', f'{lang}.diag'),
            (r'\bquad_form\b', f'{lang}.quad_form'),
            (r'\bsum_squares\b', f'{lang}.sum_squares'),
        ])

        self.tex_map = OrderedDict([
            (r'\*\*(\d+)', '^{\\1}'),
            (r'\b(\w+)\s*\*\s*(\w+)\b', r'\1 \2'),
            (r'\@', r' '),
            (r'dot', r' '),
            (r'sum_squares\((.*?)\)', r"SUM((\1))^2"),
            (r'multiply\(([^,]+), ([^)]+)\)', r'\1 \2'),
            (r'\bnp.linalg.pinv(\d+)', r'\1^{\-1}'),
            (r'\bpos\b', 'F^{+}'),
            (r'mul\((.*?),\s*(.*?)\)', r'\1 \2'),
            (r'\bmul\b\((.*?),\s*(.*?)\)', r'\1 \2'),
            (r'\bsum\b', 'SUM'),
            (r'power\((.*?),\s*(\d+)\)', r'\1^\2'),
            (r'(\w+).dual_variables\[0\]', r'\phi[\1]'),
        ])

        # mapping dict for evaluating expressions
        self.val_map = OrderedDict([
            (r'(== 0|<= 0)$', ''),  # remove the comparison operator
            (r'cp\.(Minimize|Maximize)', r'float'),  # remove cp.Minimize/Maximize
            (r'\bcp.\b', 'np.'),
            (r'\bexp\b', 'np.exp'),
            (r'\blog\b', 'np.log'),
            (r'\bconj\b', 'np.conj'),
        ])

        self.status = {
            'optimal': 0,
            'infeasible': 1,
            'unbounded': 2,
            'infeasible_inaccurate': 3,
            'unbounded_inaccurate': 4,
            'optimal_inaccurate': 5,
            'solver_error': 6,
            'time_limit': 7,
            'interrupted': 8,
            'unknown': 9,
            'infeasible_or_unbounded': 1.5,
            'user_limit': 10,
        }

    def generate_symbols(self, force_generate=False):
        """
        Generate symbols for all variables.
        """
        logger.debug(f'Entering symbol generation for <{self.parent.class_name}>')

        if (not force_generate) and self.parent._syms:
            logger.debug(' - Symbols already generated')
            return True
        t, _ = elapsed()

        # process tex_names defined in routines
        # -----------------------------------------------------------
        for key in self.parent.tex_names.keys():
            self.tex_names[key] = sp.symbols(self.parent.tex_names[key])

        # Vars
        for vname, var in self.parent.vars.items():
            self.inputs_dict[vname] = sp.symbols(f'{vname}')
            self.sub_map[rf"\b{vname}\b"] = f"self.om.{vname}"
            self.tex_map[rf"\b{vname}\b"] = rf'{var.tex_name}'
            self.val_map[rf"\b{vname}\b"] = f"rtn.{vname}.v"

        # RParams
        for rpname, rparam in self.parent.rparams.items():
            tmp = sp.symbols(f'{rparam.name}')
            self.inputs_dict[rpname] = tmp
            sub_name = ''
            if isinstance(rparam.owner, MatProcessor):
                # sparse matrices are accessed from MatProcessor
                # otherwise, dense matrices are accessed from Routine
                if rparam.sparse:
                    sub_name = f'self.rtn.system.mats.{rpname}._v'
                else:
                    sub_name = f'self.rtn.{rpname}.v'
            elif rparam.no_parse:
                sub_name = f'self.rtn.{rpname}.v'
            else:
                sub_name = f'self.om.{rpname}'
            self.sub_map[rf"\b{rpname}\b"] = sub_name
            self.tex_map[rf"\b{rpname}\b"] = f'{rparam.tex_name}'
            if not rparam.no_parse:
                self.val_map[rf"\b{rpname}\b"] = f"rtn.{rpname}.v"

        # Routine Services
        for sname, service in self.parent.services.items():
            tmp = sp.symbols(f'{service.name}')
            self.services_dict[sname] = tmp
            self.inputs_dict[sname] = tmp
            sub_name = f'self.rtn.{sname}.v' if service.no_parse else f'self.om.{sname}'
            self.sub_map[rf"\b{sname}\b"] = sub_name
            self.tex_map[rf"\b{sname}\b"] = f'{service.tex_name}'
            if not service.no_parse:
                self.val_map[rf"\b{sname}\b"] = f"rtn.{sname}.v"

        # Expressions
        for ename, expr in self.parent.exprs.items():
            self.inputs_dict[ename] = sp.symbols(f'{ename}')
            self.sub_map[rf"\b{ename}\b"] = f"self.om.{ename}"
            self.val_map[rf"\b{ename}\b"] = f"rtn.{ename}.v"
            self.tex_map[rf"\b{ename}\b"] = f'{expr.tex_name}'

        # Constraints
        # NOTE: constraints are included in sub_map for ExpressionCalc
        # thus, they don't have the suffix `.v`
        for cname, _ in self.parent.constrs.items():
            self.sub_map[rf"\b{cname}\b"] = f'self.rtn.{cname}.optz'

        # store tex names defined in `self.config`
        for key in self.config.as_dict():
            tmp = sp.symbols(key)
            self.sub_map[rf"\b{key}\b"] = f'self.rtn.config.{key}'
            if key not in self.config.tex_names.keys():
                logger.debug(f'No tex name for config.{key}')
                self.tex_map[rf"\b{key}\b"] = key
            else:
                self.tex_map[rf"\b{key}\b"] = self.config.tex_names[key]
            self.inputs_dict[key] = tmp
            if key in self.config.tex_names:
                self.tex_names[tmp] = sp.Symbol(self.config.tex_names[key])

        # store tex names for pretty printing replacement later
        for var in self.inputs_dict:
            if var in self.parent.__dict__ and self.parent.__dict__[var].tex_name is not None:
                self.tex_names[sp.symbols(var)] = sp.symbols(self.parent.__dict__[var].tex_name)

        # additional variables by conventions that are defined in ``BaseRoutine``
        self.inputs_dict['sys_f'] = sp.symbols('sys_f')
        self.inputs_dict['sys_mva'] = sp.symbols('sys_mva')

        self.parent._syms = True
        _, s = elapsed(t)

        logger.debug(f' - Symbols generated in {s}')
        return self.parent._syms

    def _check_expr_symbols(self, expr):
        """
        Check if expression contains unknown symbols.
        """
        fs = expr.free_symbols
        for item in fs:
            if item not in self.inputs_dict.values():
                raise ValueError(f'{self.class_name} expression "{expr}" contains unknown symbol "{item}"')

        fs = sorted(fs, key=lambda s: s.name)
        return fs

    def generate_pretty_print(self):
        """
        Generate pretty print math formulation.
        """
        raise NotImplementedError



"""
Module for optimization Objective.
"""
import logging

from typing import Optional
import re

import numpy as np

import cvxpy as cp

from ams.utils import pretty_long_message
from ams.shared import _prefix, _max_length

from ams.opt import OptzBase, ensure_symbols, ensure_mats_and_parsed

logger = logging.getLogger(__name__)


class Objective(OptzBase):
    """
    Base class for objective functions.

    This class serves as a template for defining objective functions. Each
    instance of this class represents a single objective function that can
    be minimized or maximized depending on the sense ('min' or 'max').

    Parameters
    ----------
    name : str, optional
        A user-defined name for the objective function.
    e_str : str, optional
        A mathematical expression representing the objective function.
    info : str, optional
        Additional informational text about the objective function.
    sense : str, optional
        The sense of the objective function, default to 'min'.
        `min` for minimization and `max` for maximization.

    Attributes
    ----------
    v : NoneType
        The value of the objective function. It needs to be set through
        computation.
    rtn : ams.routines.Routine
        The owner routine instance.
    code : str
        The code string for the objective function.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 e_str: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 sense: Optional[str] = 'min'):
        OptzBase.__init__(self, name=name, info=info, unit=unit)
        self.e_str = e_str
        self.sense = sense
        self.code = None

    @property
    def e(self):
        """
        Return the calculated objective value.

        Note that `v` should be used primarily as it is obtained
        from the solver directly.

        `e` is for debugging purpose. For a successfully solved problem,
        `e` should equal to `v`. However, when a problem is infeasible
        or unbounded, `e` can be used to check the objective value.
        """
        if self.code is None:
            logger.info(f"Objective <{self.name}> is not parsed yet.")
            return None

        val_map = self.om.rtn.syms.val_map
        code = self.code
        for pattern, replacement in val_map.items():
            try:
                code = re.sub(pattern, replacement, code)
            except TypeError as e:
                logger.error(f"Error in parsing value for obj <{self.name}>.")
                raise e

        try:
            logger.debug(pretty_long_message(f"Value code: {code}",
                                             _prefix, max_length=_max_length))
            local_vars = {'self': self, 'np': np, 'cp': cp, 'val_map': val_map}
            return self._evaluate_expression(code, local_vars)
        except Exception as e:
            logger.error(f"Error in calculating obj <{self.name}>.\n{e}")
            return None

    @property
    def v(self):
        """
        Return the CVXPY objective value.
        """
        if self.optz is None:
            return None
        else:
            return self.optz.value

    @v.setter
    def v(self, value):
        raise AttributeError("Cannot set the value of the objective function.")

    @ensure_symbols
    def parse(self):
        """
        Parse the objective function.

        Returns
        -------
        bool
            Returns True if the parsing is successful, False otherwise.
        """
        # parse the expression str
        sub_map = self.om.rtn.syms.sub_map
        code_obj = self.e_str
        for pattern, replacement, in sub_map.items():
            try:
                code_obj = re.sub(pattern, replacement, code_obj)
            except Exception as e:
                raise Exception(f"Error in parsing obj <{self.name}>.\n{e}")
        # store the parsed expression str code
        self.code = code_obj
        if self.sense not in ['min', 'max']:
            raise ValueError(f'Objective sense {self.sense} is not supported.')
        sense = 'cp.Minimize' if self.sense == 'min' else 'cp.Maximize'
        self.code = f"{sense}({code_obj})"
        msg = f" - Objective <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        return True

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the objective function.

        Returns
        -------
        bool
            Returns True if the evaluation is successful, False otherwise.
        """
        logger.debug(f" - Objective <{self.name}>: {self.e_str}")
        try:
            local_vars = {'self': self, 'cp': cp}
            self.optz = self._evaluate_expression(self.code, local_vars=local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating Objective <{self.name}>.\n{e}")
        return True

    def _evaluate_expression(self, code, local_vars=None):
        """
        Helper method to evaluate the expression code.

        Parameters
        ----------
        code : str
            The code string representing the expression.

        Returns
        -------
        cp.Expression
            The evaluated cvxpy expression.
        """
        return eval(code, {}, local_vars)

    def __repr__(self):
        return f"{self.class_name}: {self.name} [{self.sense.upper()}]"



import sys
import os
from collections import defaultdict

from setuptools import find_packages, setup

import versioneer

if sys.version_info < (3, 6):
    error = """
ams does not support Python <= {0}.{1}.
Python 3.6 and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(3, 6)
    sys.exit(error)

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as readme_file:
    readme = readme_file.read()


def parse_requires(filename):
    with open(os.path.join(here, filename)) as requirements_file:
        reqs = [
            line for line in requirements_file.read().splitlines()
            if not line.startswith('#')
        ]
    return reqs


def get_extra_requires(filename, add_all=True):
    """
    Build ``extras_require`` from an invert requirements file.

    See:
    https://hanxiao.io/2019/11/07/A-Better-Practice-for-Managing-extras-require-Dependencies-in-Python/
    """

    with open(os.path.join(here, filename)) as fp:
        extra_deps = defaultdict(set)
        for k in fp:
            if k.strip() and not k.startswith('#'):
                tags = set()
                if '#' in k:
                    if k.count("#") > 1:
                        raise ValueError("Invalid line: {}".format(k))

                    k, v = k.split('#')
                    tags.update(vv.strip() for vv in v.split(','))

                # Do not add the package name itself as a tag
                for t in tags:
                    extra_deps[t].add(k.strip())

        # add tag `all` at the end
        if add_all:
            extra_deps['all'] = set(vv for v in extra_deps.values() for vv in v)

    return extra_deps


extras_require = get_extra_requires("requirements-dev.txt")

# --- update `extras_conda` to include packages only available in PyPI ---

setup(
    name='ltbams',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Python software for scheduling modeling and co-simulation with dynanics.",
    long_description=readme,
    long_description_content_type='text/markdown',
    author="Jinning Wang",
    author_email='jinninggm@gmail.com',
    url='https://github.com/CURENT/ams',
    packages=find_packages(exclude=[]),
    entry_points={
        'console_scripts': [
            'ams = ams.cli:main',
        ],
    },
    include_package_data=True,
    package_data={
        'ltbams': [
            # When adding files here, remember to update MANIFEST.in as well,
            # or else they will not be included in the distribution on PyPI!
            # 'path/to/data_file',
        ]
    },
    install_requires=parse_requires('requirements.txt'),
    extras_require=extras_require,
    license='GNU General Public License v3 or later (GPLv3+)',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Environment :: Console',
    ],
)



import logging

from andes.core.common import Config as AndesConfig

logger = logging.getLogger(__name__)


class Config(AndesConfig):
    """
    A class for storing configuration, can be used in system,
    model, routine, and other modules.

    Revised from `andes.core.common.Config`.
    """

    def __init__(self, name, dct=None, **kwargs):
        super().__init__(name, dct, **kwargs)

    def update(self, dct: dict = None, **kwargs):
        """
        Update the configuration from a file.
        """
        if dct is not None:
            kwargs.update(dct)

        for key, val in kwargs.items():
            self._set(key, val)
            self._dict[key] = val

        self.check()



"""
Routines using PYPOWER.
"""
import logging
from typing import Optional, Union, Type
from collections import OrderedDict

from andes.shared import deg2rad, np
from andes.utils.misc import elapsed

from ams.io.pypower import system2ppc
from ams.core.param import RParam

from ams.opt import Var, Objective, ExpressionCalc
from ams.routines.routine import RoutineBase
from ams.shared import ppoption, runpf, runopf

logger = logging.getLogger(__name__)


class DCPF1(RoutineBase):
    """
    DC Power Flow using PYPOWER.

    This routine provides a wrapper for running DC power flow analysis using the
    PYPOWER.
    It leverages PYPOWER's internal DC power flow solver and maps results back to
    the AMS system.

    Notes
    -----
    - This class does not implement the AMS-style DC power flow formulation.
    - For detailed mathematical formulations and algorithmic details, refer to the
      MATPOWER User's Manual, section on Power Flow.
    """

    def __init__(self, system, config):
        RoutineBase.__init__(self, system, config)
        self.info = 'DC Power Flow'
        self.type = 'PF'

        self.map1 = OrderedDict()   # DCPF does not receive
        self.map2.update({
            'vBus': ('Bus', 'v0'),
            'ug': ('StaticGen', 'u'),
            'pg': ('StaticGen', 'p0'),
        })

        self.config.add(OrderedDict((('verbose', 1),
                                     ('out_all', 0),
                                     ('out_sys_sum', 1),
                                     ('out_area_sum', 0),
                                     ('out_bus', 1),
                                     ('out_branch', 1),
                                     ('out_gen', 0),
                                     ('out_all_lim', -1),
                                     ('out_v_lim', 1),
                                     ('out_line_lim', 1),
                                     ('out_pg_lim', 1),
                                     ('out_qg_lim', 1),
                                     )))
        self.config.add_extra("_help",
                              verbose="0: no progress info, 1: little, 2: lots, 3: all",
                              out_all="-1: individual flags control what prints, 0: none, 1: all",
                              out_sys_sum="print system summary",
                              out_area_sum="print area summaries",
                              out_bus="print bus detail",
                              out_branch="print branch detail",
                              out_gen="print generator detail (OUT_BUS also includes gen info)",
                              out_all_lim="-1: individual flags, 0: none, 1: binding, 2: all",
                              out_v_lim="0: don't print, 1: binding constraints only, 2: all constraints",
                              out_line_lim="0: don't print, 1: binding constraints only, 2: all constraints",
                              out_pg_lim="0: don't print, 1: binding constraints only, 2: all constraints",
                              out_qg_lim="0: don't print, 1: binding constraints only, 2: all constraints",
                              )
        self.config.add_extra("_alt",
                              verbose=(0, 1, 2, 3),
                              out_all=(-1, 0, 1),
                              out_sys_sum=(0, 1),
                              out_area_sum=(0, 1),
                              out_bus=(0, 1),
                              out_branch=(0, 1),
                              out_gen=(0, 1),
                              out_all_lim=(-1, 0, 1, 2),
                              out_v_lim=(0, 1, 2),
                              out_line_lim=(0, 1, 2),
                              out_pg_lim=(0, 1, 2),
                              out_qg_lim=(0, 1, 2),
                              )
        self.config.add_extra("_tex",
                              verbose=r'v_{erbose}',
                              out_all=r'o_{ut\_all}',
                              out_sys_sum=r'o_{ut\_sys\_sum}',
                              out_area_sum=r'o_{ut\_area\_sum}',
                              out_bus=r'o_{ut\_bus}',
                              out_branch=r'o_{ut\_branch}',
                              out_gen=r'o_{ut\_gen}',
                              out_all_lim=r'o_{ut\_all\_lim}',
                              out_v_lim=r'o_{ut\_v\_lim}',
                              out_line_lim=r'o_{ut\_line\_lim}',
                              out_pg_lim=r'o_{ut\_pg\_lim}',
                              out_qg_lim=r'o_{ut\_qg\_lim}',
                              )

        self.ug = RParam(info='Gen connection status',
                         name='ug', tex_name=r'u_{g}',
                         model='StaticGen', src='u',
                         no_parse=True)

        self.c2 = RParam(info='Gen cost coefficient 2',
                         name='c2', tex_name=r'c_{2}',
                         unit=r'$/(p.u.^2)', model='GCost',
                         indexer='gen', imodel='StaticGen',
                         nonneg=True, no_parse=True)
        self.c1 = RParam(info='Gen cost coefficient 1',
                         name='c1', tex_name=r'c_{1}',
                         unit=r'$/(p.u.)', model='GCost',
                         indexer='gen', imodel='StaticGen',)
        self.c0 = RParam(info='Gen cost coefficient 0',
                         name='c0', tex_name=r'c_{0}',
                         unit=r'$', model='GCost',
                         indexer='gen', imodel='StaticGen',
                         no_parse=True)

        # --- bus ---
        self.aBus = Var(info='bus voltage angle',
                        unit='rad',
                        name='aBus', tex_name=r'a_{Bus}',
                        model='Bus', src='a',)
        self.vBus = Var(info='Bus voltage magnitude',
                        unit='p.u.',
                        name='vBus', tex_name=r'v_{Bus}',
                        src='v', model='Bus',)
        # --- gen ---
        self.pg = Var(info='Gen active power',
                      unit='p.u.',
                      name='pg', tex_name=r'p_{g}',
                      model='StaticGen', src='p',)
        self.qg = Var(info='Gen reactive power',
                      unit='p.u.',
                      name='qg', tex_name=r'q_{g}',
                      model='StaticGen', src='q',)
        # --- line flow ---
        self.plf = Var(info='Line flow',
                       unit='p.u.',
                       name='plf', tex_name=r'p_{lf}',
                       model='Line',)
        # --- objective ---
        self.obj = Objective(name='obj',
                             info='total cost',
                             e_str='0',
                             sense='min',)

        # --- total cost ---
        tcost = 'sum(mul(c2, pg**2))'
        tcost += '+ sum(mul(c1, pg))'
        tcost += '+ sum(mul(ug, c0))'
        self.tcost = ExpressionCalc(info='Total cost', unit='$',
                                    model=None, src=None,
                                    e_str=tcost)

    def solve(self, **kwargs):
        """
        Solve by PYPOWER.
        """
        ppc = system2ppc(self.system)
        config = {key.upper(): value for key, value in self.config._dict.items()}
        # Enforece DC power flow
        ppopt = ppoption(PF_DC=True, **config)
        res, _ = runpf(casedata=ppc, ppopt=ppopt)
        return res

    def unpack(self, res, **kwargs):
        """
        Unpack results from PYPOWER.
        """
        system = self.system
        mva = res['baseMVA']

        # --- copy results from ppc into system algeb ---
        # --- Bus ---
        system.Bus.v.v = res['bus'][:, 7]               # voltage magnitude
        system.Bus.a.v = res['bus'][:, 8] * deg2rad     # voltage angle

        # --- PV ---
        system.PV.p.v = res['gen'][system.Slack.n:, 1] / mva        # active power
        system.PV.q.v = res['gen'][system.Slack.n:, 2] / mva        # reactive power

        # --- Slack ---
        system.Slack.p.v = res['gen'][:system.Slack.n, 1] / mva     # active power
        system.Slack.q.v = res['gen'][:system.Slack.n, 2] / mva     # reactive power

        # --- Line ---
        self.plf.optz.value = res['branch'][:, 13] / mva  # line flow

        # NOTE: In PYPOWER, branch status is not optimized and this assignment
        # typically has no effect on results. However, in some extensions (e.g., gurobi-optimods),
        # branch status may be optimized. This line ensures that the system's branch status
        # is updated to reflect the results from the solver, if applicable.

        system.Line.u.v = res['branch'][:, 10]

        # --- copy results from system algeb into routine algeb ---
        for vname, var in self.vars.items():
            owner = getattr(system, var.model)  # instance of owner, Model or Group
            if var.src is None:          # skip if no source variable is specified
                continue
            elif hasattr(owner, 'group'):   # if owner is a Model instance
                grp = getattr(system, owner.group)
                idx = grp.get_all_idxes()
            elif hasattr(owner, 'get_idx'):  # if owner is a Group instance
                idx = owner.get_all_idxes()
            else:
                msg = f"Failed to find valid source variable `{owner.class_name}.{var.src}` for "
                msg += f"{self.class_name}.{vname}, skip unpacking."
                logger.warning(msg)
                continue
            try:
                logger.debug(f"Unpacking {vname} into {owner.class_name}.{var.src}.")
                var.optz.value = owner.get(src=var.src, attr='v', idx=idx)
            except AttributeError:
                logger.debug(f"Failed to unpack {vname} into {owner.class_name}.{var.src}.")
                continue
        self.system.recent = self.system.routines[self.class_name]
        return True

    def run(self, **kwargs):
        """
        Run the DC power flow using PYPOWER.

        Returns
        -------
        bool
            True if the optimization converged successfully, False otherwise.
        """
        if not self.initialized:
            self.init()
        t0, _ = elapsed()

        # --- solve optimization ---
        t0, _ = elapsed()
        res = self.solve(**kwargs)
        self.converged = res['success']
        self.exit_code = 0 if res['success'] else 1
        _, s = elapsed(t0)
        self.exec_time = float(s.split(" ")[0])
        try:
            n_iter = res['raw']['output']['iterations']
        except Exception:
            n_iter = -1
        n_iter_str = f"{n_iter} iterations " if n_iter > 1 else f"{n_iter} iteration "
        if self.exit_code == 0:
            msg = f"<{self.class_name}> converged in {s}, "
            msg += n_iter_str + "with PYPOWER."
            logger.warning(msg)
            try:
                self.unpack(res)
            except Exception as e:
                logger.error(f"Failed to unpack results from {self.class_name}.\n{e}")
                return False
            self.system.report()
            return True
        else:
            msg = f"{self.class_name} failed to converge in {s}, "
            msg += n_iter_str + "with PYPOWER."
            logger.warning(msg)
            return False

    def _get_off_constrs(self):
        pass

    def _data_check(self, info=True, **kwargs):
        pass

    def update(self, params=None, build_mats=False, **kwargs):
        pass

    def enable(self, name):
        raise NotImplementedError

    def disable(self, name):
        raise NotImplementedError

    def _post_add_check(self):
        pass

    def addRParam(self,
                  name: str,
                  tex_name: Optional[str] = None,
                  info: Optional[str] = None,
                  src: Optional[str] = None,
                  unit: Optional[str] = None,
                  model: Optional[str] = None,
                  v: Optional[np.ndarray] = None,
                  indexer: Optional[str] = None,
                  imodel: Optional[str] = None,):
        raise NotImplementedError

    def addService(self,
                   name: str,
                   value: np.ndarray,
                   tex_name: str = None,
                   unit: str = None,
                   info: str = None,
                   vtype: Type = None,):
        raise NotImplementedError

    def addConstrs(self,
                   name: str,
                   e_str: str,
                   info: Optional[str] = None,
                   is_eq: Optional[str] = False,):
        raise NotImplementedError

    def addVars(self,
                name: str,
                model: Optional[str] = None,
                shape: Optional[Union[int, tuple]] = None,
                tex_name: Optional[str] = None,
                info: Optional[str] = None,
                src: Optional[str] = None,
                unit: Optional[str] = None,
                horizon: Optional[RParam] = None,
                nonneg: Optional[bool] = False,
                nonpos: Optional[bool] = False,
                cplx: Optional[bool] = False,
                imag: Optional[bool] = False,
                symmetric: Optional[bool] = False,
                diag: Optional[bool] = False,
                psd: Optional[bool] = False,
                nsd: Optional[bool] = False,
                hermitian: Optional[bool] = False,
                boolean: Optional[bool] = False,
                integer: Optional[bool] = False,
                pos: Optional[bool] = False,
                neg: Optional[bool] = False,):
        raise NotImplementedError


class PFlow1(DCPF1):
    """
    Power Flow using PYPOWER.

    This routine provides a wrapper for running power flow analysis using the
    PYPOWER.
    It leverages PYPOWER's internal power flow solver and maps results back to the
    AMS system.

    Known Issues
    ------------
    - Fast-Decoupled (XB version) and Fast-Decoupled (BX version) algorithms are
      not fully supported yet.

    Notes
    -----
    - This class does not implement the AMS-style power flow formulation.
    - For detailed mathematical formulations and algorithmic details, refer to the
      MATPOWER User's Manual, section on Power Flow.
    """

    def __init__(self, system, config):
        DCPF1.__init__(self, system, config)
        self.info = 'Power Flow'
        self.type = 'PF'

        # PFlow does not receive nor send
        self.map1 = OrderedDict()
        self.map2 = OrderedDict()

        self.config.add(OrderedDict((('pf_alg', 1),
                                     ('pf_tol', 1e-8),
                                     ('pf_max_it', 10),
                                     ('pf_max_it_fd', 30),
                                     ('pf_max_it_gs', 1000),
                                     ('enforce_q_lims', 0),
                                     )))
        self.config.add_extra("_help",
                              pf_alg="1: Newton, 2: Fast-Decoupled XB, 3: Fast-Decoupled BX, 4: Gauss Seidel",
                              pf_tol="termination tolerance on per unit P & Q mismatch",
                              pf_max_it="maximum number of iterations for Newton's method",
                              pf_max_it_fd="maximum number of iterations for fast decoupled method",
                              pf_max_it_gs="maximum number of iterations for Gauss-Seidel method",
                              enforce_q_lims="enforce gen reactive power limits, at expense of |V|",
                              )
        self.config.add_extra("_alt",
                              pf_alg=(1, 2, 3, 4),
                              pf_tol=(0.0, 1e-8),
                              pf_max_it=">1",
                              pf_max_it_fd=">1",
                              pf_max_it_gs=">1",
                              enforce_q_lims=(0, 1),
                              )

    def solve(self, **kwargs):
        ppc = system2ppc(self.system)
        config = {key.upper(): value for key, value in self.config._dict.items()}
        # Enforece AC power flow
        ppopt = ppoption(PF_DC=False, **config)
        res, _ = runpf(casedata=ppc, ppopt=ppopt)
        return res

    def run(self, **kwargs):
        """
        Run the power flow using PYPOWER.

        Returns
        -------
        bool
            True if the optimization converged successfully, False otherwise.
        """
        return super().run(**kwargs)


class DCOPF1(DCPF1):
    """
    DC optimal power flow using PYPOWER.

    This routine provides a wrapper for running DC optimal power flow analysis using
    the PYPOWER.
    It leverages PYPOWER's internal DC optimal power flow solver and maps results
    back to the AMS system.

    In PYPOWER, the ``c0`` term (the constant coefficient in the generator cost
    function) is always included in the objective, regardless of the generator's
    commitment status. See `pypower/opf_costfcn.py` for implementation details.

    Known Issues
    ------------
    - Algorithms 400, 500, 600, and 700 are not fully supported yet.

    Notes
    -----
    - This class does not implement the AMS-style DC optimal power flow formulation.
    - For detailed mathematical formulations and algorithmic details, refer to the
      MATPOWER User's Manual, section on Optimal Power Flow.
    """

    def __init__(self, system, config):
        DCPF1.__init__(self, system, config)
        self.info = 'DC Optimal Power Flow'
        self.type = 'DCED'

        self.map1 = OrderedDict()   # DCOPF does not receive
        self.map2.update({
            'vBus': ('Bus', 'v0'),
            'ug': ('StaticGen', 'u'),
            'pg': ('StaticGen', 'p0'),
        })
        self.config.add(OrderedDict((('opf_alg_dc', 200),
                                     ('opf_violation', 5e-6),
                                     ('opf_flow_lim', 0),
                                     ('opf_ignore_ang_lim', 0),
                                     ('grb_method', 1),
                                     ('grb_timelimit', float('inf')),
                                     ('grb_threads', 0),
                                     ('grb_opt', 0),
                                     ('pdipm_feastol', 0),
                                     ('pdipm_gradtol', 1e-6),
                                     ('pdipm_comptol', 1e-6),
                                     ('pdipm_costtol', 1e-6),
                                     ('pdipm_max_it', 150),
                                     ('scpdipm_red_it', 20),
                                     )))
        opf_alg_dc = "0: choose default solver based on availability, 200: PIPS, 250: PIPS-sc, "
        opf_alg_dc += "400: IPOPT, 500: CPLEX, 600: MOSEK, 700: GUROBI"
        opf_flow_lim = "qty to limit for branch flow constraints: 0 - apparent power flow (limit in MVA), "
        opf_flow_lim += "1 - active power flow (limit in MW), "
        opf_flow_lim += "2 - current magnitude (limit in MVA at 1 p.u. voltage)"
        grb_method = "0 - primal simplex, 1 - dual simplex, 2 - barrier, 3 - concurrent (LP only), "
        grb_method += "4 - deterministic concurrent (LP only)"
        pdipm_feastol = "feasibility (equality) tolerance for Primal-Dual Interior Points Methods, "
        pdipm_feastol += "set to value of OPF_VIOLATION by default"
        pdipm_gradtol = "gradient tolerance for Primal-Dual Interior Points Methods"
        pdipm_comptol = "complementary condition (inequality) tolerance for Primal-Dual Interior Points Methods"
        scpdipm_red_it = "maximum reductions per iteration for Step-Control Primal-Dual Interior Points Methods"
        self.config.add_extra("_help",
                              opf_alg_dc=opf_alg_dc,
                              opf_violation="constraint violation tolerance",
                              opf_flow_lim=opf_flow_lim,
                              opf_ignore_ang_lim="ignore angle difference limits for branches even if specified",
                              grb_method=grb_method,
                              grb_timelimit="maximum time allowed for solver (TimeLimit)",
                              grb_threads="(auto) maximum number of threads to use (Threads)",
                              grb_opt="See gurobi_options() for details",
                              pdipm_feastol=pdipm_feastol,
                              pdipm_gradtol=pdipm_gradtol,
                              pdipm_comptol=pdipm_comptol,
                              pdipm_costtol="optimality tolerance for Primal-Dual Interior Points Methods",
                              pdipm_max_it="maximum iterations for Primal-Dual Interior Points Methods",
                              scpdipm_red_it=scpdipm_red_it,
                              )
        self.config.add_extra("_alt",
                              opf_alg_dc=(0, 200, 250, 400, 500, 600, 700),
                              opf_violation=">=0",
                              opf_flow_lim=(0, 1, 2),
                              opf_ignore_ang_lim=(0, 1),
                              grb_method=(0, 1, 2, 3, 4),
                              grb_timelimit=(0, float('inf')),
                              grb_threads=(0, 1),
                              grb_opt=(0, 1),
                              pdipm_feastol=">=0",
                              pdipm_gradtol=">=0",
                              pdipm_comptol=">=0",
                              pdipm_costtol=">=0",
                              pdipm_max_it=">=0",
                              scpdipm_red_it=">=0",
                              )
        self.config.add_extra("_tex",
                              opf_alg_dc=r'o_{pf\_alg\_dc}',
                              opf_violation=r'o_{pf\_violation}',
                              opf_flow_lim=r'o_{pf\_flow\_lim}',
                              opf_ignore_ang_lim=r'o_{pf\_ignore\_ang\_lim}',
                              grb_method=r'o_{grb\_method}',
                              grb_timelimit=r'o_{grb\_timelimit}',
                              grb_threads=r'o_{grb\_threads}',
                              grb_opt=r'o_{grb\_opt}',
                              pdipm_feastol=r'o_{pdipm\_feastol}',
                              pdipm_gradtol=r'o_{pdipm\_gradtol}',
                              pdipm_comptol=r'o_{pdipm\_comptol}',
                              pdipm_costtol=r'o_{pdipm\_costtol}',
                              pdipm_max_it=r'o_{pdipm\_max\_it}',
                              scpdipm_red_it=r'o_{scpdipm\_red\_it}',
                              )

        self.obj = Objective(name='obj',
                             info='total cost, placeholder',
                             e_str='sum(c2 * pg**2 + c1 * pg + c0)',
                             sense='min',)

        self.pi = Var(info='Lagrange multiplier on real power mismatch',
                      name='pi', unit='$/p.u.',
                      model='Bus', src=None,)
        self.piq = Var(info='Lagrange multiplier on reactive power mismatch',
                       name='piq', unit='$/p.u.',
                       model='Bus', src=None,)

        self.mu1 = Var(info='Kuhn-Tucker multiplier on MVA limit at bus1',
                       name='mu1', unit='$/p.u.',
                       model='Line', src=None,)
        self.mu2 = Var(info='Kuhn-Tucker multiplier on MVA limit at bus2',
                       name='mu2', unit='$/p.u.',
                       model='Line', src=None,)

    def solve(self, **kwargs):
        ppc = system2ppc(self.system)
        config = {key.upper(): value for key, value in self.config._dict.items()}
        ppopt = ppoption(PF_DC=True, **config)  # Enforce DCOPF
        res = runopf(casedata=ppc, ppopt=ppopt)
        return res

    def unpack(self, res, **kwargs):
        mva = res['baseMVA']
        self.pi.optz.value = res['bus'][:, 13] / mva
        self.piq.optz.value = res['bus'][:, 14] / mva
        self.mu1.optz.value = res['branch'][:, 17] / mva
        self.mu2.optz.value = res['branch'][:, 18] / mva
        return super().unpack(res)

    def run(self, **kwargs):
        """
        Run the DCOPF routine using PYPOWER.

        Returns
        -------
        bool
            True if the optimization converged successfully, False otherwise.
        """
        return super().run(**kwargs)


class ACOPF1(DCOPF1):
    """
    AC optimal power flow using PYPOWER.

    This routine provides a wrapper for running AC optimal power flow analysis using
    the PYPOWER.
    It leverages PYPOWER's internal AC optimal power flow solver and maps results
    back to the AMS system.

    In PYPOWER, the ``c0`` term (the constant coefficient in the generator cost
    function) is always included in the objective, regardless of the generator's
    commitment status. See `pypower/opf_costfcn.py` for implementation details.

    Notes
    -----
    - This class does not implement the AMS-style AC optimal power flow formulation.
    - For detailed mathematical formulations and algorithmic details, refer to the
      MATPOWER User's Manual, section on Optimal Power Flow.
    """

    def __init__(self, system, config):
        DCOPF1.__init__(self, system, config)
        self.info = 'AC Optimal Power Flow'
        self.type = 'ACED'

        self.map1 = OrderedDict()   # ACOPF does not receive
        self.map2.update({
            'vBus': ('Bus', 'v0'),
            'ug': ('StaticGen', 'u'),
            'pg': ('StaticGen', 'p0'),
        })

        self.config.add(OrderedDict((('opf_alg', 0),
                                     )))
        self.config.add_extra("_help",
                              opf_alg="algorithm to use for OPF: 0 - default, 580 - PIPS"
                              )
        self.config.add_extra("_alt",
                              opf_alg=(0, 580),
                              )
        self.config.add_extra("_tex",
                              opf_alg=r'o_{pf\_alg}',
                              )

    def solve(self, **kwargs):
        ppc = system2ppc(self.system)
        config = {key.upper(): value for key, value in self.config._dict.items()}
        ppopt = ppoption(PF_DC=False, **config)
        res = runopf(casedata=ppc, ppopt=ppopt)
        return res

    def run(self, **kwargs):
        """
        Run the ACOPF routine using PYPOWER.

        Returns
        -------
        bool
            True if the optimization converged successfully, False otherwise.
        """
        super().run(**kwargs)



"""
Module for the information container models.
"""
from andes.core.param import DataParam
from andes.core.model.modeldata import ModelData

from ams.core.model import Model


class Summary(ModelData, Model):
    """
    Class for storing system summary.
    Can be used for random information or notes.
    """

    def __init__(self, system, config):
        ModelData.__init__(self, three_params=False)

        self.field = DataParam(info='field name')
        self.comment = DataParam(info='information, comment, or anything')
        self.comment2 = DataParam(info='comment field 2')
        self.comment3 = DataParam(info='comment field 3')
        self.comment4 = DataParam(info='comment field 4')

        Model.__init__(self, system, config)
        self.group = 'Information'



"""
Module for report generation.
"""
import logging
from collections import OrderedDict
from typing import List, Dict, Optional

from andes.io.txt import dump_data
from andes.shared import np
from andes.utils.misc import elapsed

from ams import __version__ as version
from ams.shared import copyright_msg, nowarranty_msg, report_time

logger = logging.getLogger(__name__)


DECIMALS = 6


def report_info(system) -> list:
    info = list()
    info.append('AMS' + ' ' + version + '\n')
    info.append(f'{copyright_msg}\n\n')
    info.append(f"{nowarranty_msg}\n")
    info.append('Case file: ' + str(system.files.case) + '\n')
    info.append(f'Report time: {report_time}\n\n')
    return info


class Report:
    """
    Report class to store routine analysis reports.

    Notes
    -----
    Revised from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    def __init__(self, system):
        self.system = system
        self.basic = OrderedDict()

    @property
    def info(self):
        return report_info(self.system)

    def update(self):
        """
        Update values based on the requested content
        """
        system = self.system
        self.basic.update({
            'Buses': system.Bus.n,
            'Generators': system.PV.n + system.Slack.n,
            'Loads': system.PQ.n,
            'Shunts': system.Shunt.n,
            'Lines': system.Line.n,
            'Transformers': np.count_nonzero(system.Line.trans.v == 1),
            'Areas': system.Area.n,
            'Zones': system.Zone.n,
        })

    def collect(self, rtn, horizon=None):
        """
        Collect report data.

        Parameters
        ----------
        rtn : Routine
            Routine object to collect data from.
        horizon : str, optional
            Timeslot to collect data from. Only single timeslot is supported.
        """
        text = list()
        header = list()
        row_name = list()
        data = list()

        if not rtn.converged:
            return text, header, row_name, data

        owners = collect_owners(rtn)
        owners = collect_vars(owners, rtn, horizon, DECIMALS)
        owners = collect_exprs(owners, rtn, horizon, DECIMALS)
        owners = collect_exprcs(owners, rtn, horizon, DECIMALS)

        dump_collected_data(owners, text, header, row_name, data)

        return text, header, row_name, data

    def write(self):
        """
        Write report to file.
        """
        system = self.system
        if system.files.no_output is True:
            return

        text = list()
        header = list()
        row_name = list()
        data = list()
        self.update()

        t, _ = elapsed()

        # --- system info section ---
        text.append(self.info)
        header.append(None)
        row_name.append(None)
        data.append(None)

        # --- system summary section ---
        text.append(['='*10 + ' System Statistics ' + '='*10 + '\n'])
        header.append(None)
        row_name.append(self.basic.keys())
        data.append(list(self.basic.values()))

        # --- rountine data section ---
        rtns_to_collect = [rtn for rtn in system.routines.values() if rtn.converged]
        for rtn in rtns_to_collect:
            # --- routine summary ---
            text.append(['='*30 + f' {rtn.class_name} ' + '='*30])
            header.append(None)
            row_name.append(None)
            data.append(None)
            if hasattr(rtn, 'timeslot'):
                for slot in rtn.timeslot.v:
                    # --- timeslot summary ---
                    text.append(['-'*28 + f' {slot} ' + '-'*28])
                    header.append(None)
                    row_name.append(None)
                    data.append(None)
                    text_sum, header_sum, row_name_sum, data_sum = self.collect(rtn, horizon=[slot])
                    # --- timeslot data ---
                    text.extend(text_sum)
                    header.extend(header_sum)
                    row_name.extend(row_name_sum)
                    data.extend(data_sum)
            else:
                # single-period
                text_sum, header_sum, row_name_sum, data_sum = self.collect(rtn)
                # --- routine extended ---
                text.append([''])
                row_name.append(
                    ['Generation', 'Load'])

                if hasattr(rtn, 'pd'):
                    pd = rtn.pd.v.sum().round(DECIMALS)
                else:
                    pd = rtn.system.PQ.p0.v.sum().round(DECIMALS)
                if hasattr(rtn, 'qd'):
                    qd = rtn.qd.v.sum().round(DECIMALS)
                else:
                    qd = rtn.system.PQ.q0.v.sum().round(DECIMALS)

                if not hasattr(rtn, 'qg'):
                    header.append(['P (p.u.)'])
                    Pcol = [rtn.pg.v.sum().round(DECIMALS), pd]
                    data.append([Pcol])
                else:
                    header.append(['P (p.u.)', 'Q (p.u.)'])
                    Pcol = [rtn.pg.v.sum().round(DECIMALS), pd]
                    Qcol = [rtn.qg.v.sum().round(DECIMALS), qd]
                    data.append([Pcol, Qcol])

                # --- routine data ---
                text.extend(text_sum)
                header.extend(header_sum)
                row_name.extend(row_name_sum)
                data.extend(data_sum)
        dump_data(text, header, row_name, data, system.files.txt)

        _, s = elapsed(t)
        logger.info(f'Report saved to "{system.files.txt}" in {s}.')


def dump_collected_data(owners: dict, text: List, header: List, row_name: List, data: List) -> None:
    """
    Dump collected data into the provided lists.

    Parameters
    ----------
    owners : dict
        Dictionary of owners.
    text : list
        List to append text data to.
    header : list
        List to append header data to.
    row_name : list
        List to append row names to.
    data : list
        List to append data to.
    """
    for key, val in owners.items():
        text.append([f'{key} DATA:\n'])
        row_name.append(val['idx'])
        header.append(val['header'])
        data.append(val['data'])


def collect_exprcs(owners: Dict, rtn, horizon: Optional[str], decimals: int) -> Dict:
    """
    Collect expression calculations and populate the data dictionary.

    Parameters
    ----------
    owners : dict
        Dictionary of owners.
    rtn : Routine
        Routine object to collect data from.
    horizon : str, optional
        Timeslot to collect data from. Only single timeslot is supported.
    decimals : int
        Number of decimal places to round the data.

    Returns
    -------
    dict
        Updated dictionary of owners with collected ExpressionCalc data.
    """
    for key, exprc in rtn.exprcs.items():
        if exprc.owner is None:
            continue
        owner_name = exprc.owner.class_name
        idx_v = owners[owner_name]['idx']
        header_v = key if exprc.unit is None else f'{key} ({exprc.unit})'
        try:
            data_v = rtn.get(src=key, attr='v', idx=idx_v, horizon=horizon).round(decimals)
        except Exception:
            data_v = [np.nan] * len(idx_v)
        owners[owner_name]['header'].append(header_v)
        owners[owner_name]['data'].append(data_v)

    return owners


def collect_exprs(owners: Dict, rtn, horizon: Optional[str], decimals: int) -> Dict:
    """
    Collect expressions and populate the data dictionary.

    Parameters
    ----------
    owners : dict
        Dictionary of owners.
    rtn : Routine
        Routine object to collect data from.
    horizon : str, optional
        Timeslot to collect data from. Only single timeslot is supported.
    decimals : int
        Number of decimal places to round the data.

    Returns
    -------
    dict
        Updated dictionary of owners with collected Expression data.
    """
    for key, expr in rtn.exprs.items():
        if expr.owner is None:
            continue
        owner_name = expr.owner.class_name
        idx_v = owners[owner_name]['idx']
        header_v = key if expr.unit is None else f'{key} ({expr.unit})'
        try:
            data_v = rtn.get(src=key, attr='v', idx=idx_v, horizon=horizon).round(decimals)
        except Exception:
            data_v = [np.nan] * len(idx_v)
        owners[owner_name]['header'].append(header_v)
        owners[owner_name]['data'].append(data_v)

    return owners


def collect_vars(owners: Dict, rtn, horizon: Optional[str], decimals: int) -> Dict:
    """
    Collect variables and populate the data dictionary.

    Parameters
    ----------
    owners : dict
        Dictionary of owners.
    rtn : Routine
        Routine object to collect data from.
    horizon : str, optional
        Timeslot to collect data from. Only single timeslot is supported.
    decimals : int
        Number of decimal places to round the data.

    Returns
    -------
    dict
        Updated dictionary of owners with collected Var data.
    """

    for key, var in rtn.vars.items():
        if var.owner is None:
            continue
        owner_name = var.owner.class_name
        idx_v = owners[owner_name]['idx']
        header_v = key if var.unit is None else f'{key} ({var.unit})'
        try:
            data_v = rtn.get(src=key, attr='v', idx=idx_v, horizon=horizon).round(decimals)
        except Exception:
            data_v = [np.nan] * len(idx_v)
        owners[owner_name]['header'].append(header_v)
        owners[owner_name]['data'].append(data_v)

    return owners


def collect_owners(rtn):
    """
    Initialize an owners dictionary for data collection.

    Returns
    -------
    dict
        A dictionary of initialized owners.
    """
    # initialize data section by model
    owners_all = ['Bus', 'Line', 'StaticGen',
                  'PV', 'Slack', 'RenGen',
                  'DG', 'ESD1', 'PVD1', 'VSG',
                  'StaticLoad']

    # Filter owners that exist in the system
    owners_e = list({
        var.owner.class_name for var in rtn.vars.values() if var.owner is not None
    }.union(
        expr.owner.class_name for expr in rtn.exprs.values() if expr.owner is not None
    ).union(
        exprc.owner.class_name for exprc in rtn.exprcs.values() if exprc.owner is not None
    ))

    # Use a dictionary comprehension to create vars_by_owner
    owners = {
        name: {'idx': [],
               'name': [],
               'header': [],
               'data': [], }
        for name in owners_all if name in owners_e and getattr(rtn.system, name).n > 0
    }

    for key, val in owners.items():
        owner = getattr(rtn.system, key)
        idx_v = owner.get_all_idxes()
        val['idx'] = idx_v
        val['name'] = owner.get(src='name', attr='v', idx=idx_v)
        val['header'].append('Name')
        val['data'].append(val['name'])

    return owners
